/* Copyright 1999 by the Massachusetts Institute of Technology.
 *
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies and that both that copyright
 * notice and this permission notice appear in supporting
 * documentation, and that the name of M.I.T. not be used in
 * advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 * M.I.T. makes no representations about the suitability of
 * this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 */

static const char rcsid[] = "$Id: mux.c,v 1.2 1999/03/29 15:56:24 tytso Exp $";

#include <sys/types.h>
#include <sys/time.h>
#include <stdlib.h>
#include <time.h>
#include <errno.h>
#include "mux.h"

#define HEAP_START 1024

#define MUX_FLAG_BREAK	1

struct mux_data {
  mux_fd inputs;
  mux_fd outputs;
  mux_timer *timers;
  int num_timers;
  int timers_size;
  int flags;
};

struct mux_fd_data {
  int fd;
  mux_callback callback;
  void *arg;
  mux_fd next;
  mux_fd *prevp;
};

struct mux_timer_data {
  int heap_pos;
  struct timeval abstime;
  mux_callback callback;
  void *arg;
};

static void remove_timer(mux m, mux_timer timer);
static int parent(int n);
static int child1(int n);
static int child2(int n);
static void hassign(mux_timer *heap, int pos, mux_timer timer);
static int tcomp(mux_timer t1, mux_timer t2);
static int hcomp(mux_timer *heap, int pos1, int pos2);
static int tvcomp(struct timeval *tv1, struct timeval *tv2);
static void tvsub(struct timeval *result, struct timeval *tv1,
		  struct timeval *tv2);

/* The mux timers are organized into a heap.  The heap (m->timers) is
 * an array of pointers to timers, and each timer knows the position
 * of its pointer in the heap (timer->heap_pos).
 *
 * A heap is a binary tree represented as an array.  If i is some
 * position in the heap, then floor(i/2) is the parent of i and i*2+1
 * and i*2+2 are its children.  The timer heap has the invariant
 * condition that, for each position i, the timeout at i is no later
 * than the timeout at child positions i*2+1 and i*2+2.  Obviously,
 * then, the earliest timeout value must be at position 0.
 *
 * To add a timer to the heap, we start by adding it to the end, and
 * then keep swapping it with its parent until it has a parent with a
 * timer value less than its value.  This operation preserves the heap
 * property on all indices of the array.
 *
 * To delete a timer at position i from the heap, we discard it and
 * fill in its position with the last timer in the heap.  In order to
 * restore the heap property at all indices of the array, we have to
 * consider two cases: the timeout at i is less than that of its
 * parent, or the timeout at i is greater than that of one of its
 * children.  In the first case, we propagate the timer at i up the
 * tree, swapping it with its parent, until the heap is restored; in
 * the second case, we propagate the timer down the tree, swapping it
 * with its earliest-timeout child, until the heap is restored.
 */

int mux_init(mux *mptr)
{
  mux m;

  m = malloc(sizeof(struct mux_data));
  if (!m)
    return MUX_ENOMEM;
  m->inputs = NULL;
  m->outputs = NULL;
  m->timers_size = HEAP_START;
  m->timers = malloc(m->timers_size * sizeof(mux_timer));
  if (!m->timers)
    {
      free(m);
      return MUX_ENOMEM;
    }
  m->num_timers = 0;
  m->flags = 0;
  *mptr = m;
  return MUX_SUCCESS;
}

void mux_destroy(mux m)
{
  mux_fd mfd, next;
  int i;

  for (mfd = m->inputs; mfd; mfd = next)
    {
      mfd->callback(mfd->arg, MUX_EDESTRUCTION);
      next = mfd->next;
      free(mfd);
    }
  for (mfd = m->outputs; mfd; mfd = mfd->next)
    {
      mfd->callback(mfd->arg, MUX_EDESTRUCTION);
      next = mfd->next;
      free(mfd);
    }
  for (i = 0; i < m->num_timers; i++)
    {
      m->timers[i]->callback(m->timers[i]->arg, MUX_EDESTRUCTION);
      free(m->timers[i]);
    }
  free(m->timers);
  free(m);
}

void mux_break(mux m)
{
  m->flags |= MUX_FLAG_BREAK;
}

int mux_add_input(mux m, mux_fd *mfdptr, int fd, mux_callback callback,
		  void *arg)
{
  mux_fd mfd;

  mfd = malloc(sizeof(struct mux_fd_data));
  if (!mfd)
    return MUX_ENOMEM;
  mfd->fd = fd;
  mfd->callback = callback;
  mfd->arg = arg;
  mfd->next = m->inputs;
  mfd->prevp = &m->inputs;
  m->inputs = mfd;
  if (mfd->next)
    mfd->next->prevp = &mfd->next;
  if (mfdptr)
    *mfdptr = mfd;
  return MUX_SUCCESS;
}

int mux_add_output(mux m, mux_fd *mfdptr, int fd, mux_callback callback,
		   void *arg)
{
  mux_fd mfd;

  mfd = malloc(sizeof(struct mux_fd_data));
  if (!mfd)
    return MUX_ENOMEM;
  mfd->fd = fd;
  mfd->callback = callback;
  mfd->arg = arg;
  mfd->next = m->outputs;
  mfd->prevp = &m->outputs;
  m->outputs = mfd;
  if (mfd->next)
    mfd->next->prevp = &mfd->next;
  if (mfdptr)
    *mfdptr = mfd;
  return MUX_SUCCESS;
}

int mux_add_timer(mux m, mux_timer *timerptr, struct timeval *abstime,
		  mux_callback callback, void *arg)
{
  mux_timer timer, *new_heap;
  int pos, ppos;

  /* Expand the timer heap if necessary. */
  if (m->num_timers == m->timers_size)
    {
      new_heap = realloc(m->timers, m->timers_size * 2 * sizeof(mux_timer));
      if (!new_heap)
	return MUX_ENOMEM;
      m->timers = new_heap;
      m->timers_size *= 2;
    }

  /* Create the new timer object. */
  timer = malloc(sizeof(struct mux_timer_data));
  if (!timer)
    return MUX_ENOMEM;
  timer->abstime = *abstime;
  timer->callback = callback;
  timer->arg = arg;

  /* Insert the new timer into the heap. */
  pos = m->num_timers;
  while (pos > 0)
    {
      ppos = parent(pos);
      if (tcomp(timer, m->timers[ppos]) >= 0)
	break;
      hassign(m->timers, pos, m->timers[ppos]);
      pos = ppos;
    }
  hassign(m->timers, pos, timer);
  m->num_timers++;

  if (timerptr)
    *timerptr = timer;
  return MUX_SUCCESS;
}

int mux_add_rstimer(mux m, mux_timer *timerptr, time_t offset,
		    mux_callback callback, void *arg)
{
  struct timeval tv;

  gettimeofday(&tv, NULL);
  tv.tv_sec += offset;
  return mux_add_timer(m, timerptr, &tv, callback, arg);
}

void mux_cancel_fd(mux m, mux_fd mfd)
{
  mfd->callback(mfd->arg, MUX_EDESTRUCTION);
  *mfd->prevp = mfd->next;
  if (mfd->next)
    mfd->next->prevp = mfd->prevp;
  free(mfd);
}

void mux_cancel_timer(mux m, mux_timer timer)
{
  timer->callback(timer->arg, MUX_EDESTRUCTION);
  remove_timer(m, timer);
}

int mux_loop(mux m)
{
  fd_set readers, writers;
  struct timeval tv, *timeout;
  int nfds, count;

  while (1)
    {
      FD_ZERO(&readers);
      FD_ZERO(&writers);
      if (!m->inputs && !m->outputs && !m->num_timers)
        break;
      nfds = mux_fds(m, &readers, &writers);
      timeout = mux_timeout(m, NULL, &tv);
      count = select(nfds, &readers, &writers, NULL, timeout);
      if (count < 0 && errno != EINTR)
	return MUX_ESELECT;
      mux_process(m, &readers, &writers);
      if (m->flags & MUX_FLAG_BREAK)
        break;
    }
  m->flags &= ~MUX_FLAG_BREAK;
  return MUX_SUCCESS;
}

int mux_fds(mux m, fd_set *read_fds, fd_set *write_fds)
{
  mux_fd mfd;
  int nfds = 0;

  for (mfd = m->inputs; mfd; mfd = mfd->next)
    {
      FD_SET(mfd->fd, read_fds);
      if (mfd->fd >= nfds)
	nfds = mfd->fd + 1;
    }
  for (mfd = m->outputs; mfd; mfd = mfd->next)
    {
      FD_SET(mfd->fd, write_fds);
      if (mfd->fd >= nfds)
	nfds = mfd->fd + 1;
    }
  return nfds;
}

struct timeval *mux_timeout(mux m, struct timeval *maxtv, struct timeval *tv)
{
  struct timeval now;

  if (m->num_timers)
    {
      gettimeofday(&now, NULL);
      tvsub(tv, &m->timers[0]->abstime, &now);
      return (!maxtv || tvcomp(tv, maxtv) < 0) ? tv : maxtv;
    }
  else
    return maxtv;
}

void mux_process(mux m, fd_set *read_fds, fd_set *write_fds)
{
  mux_fd mfd, next;
  mux_callback callback;
  void *arg;
  struct timeval now;

  /* Process timers which have timed out. */
  gettimeofday(&now, NULL);
  while (m->num_timers > 0 && tvcomp(&m->timers[0]->abstime, &now) <= 0)
    {
      callback = m->timers[0]->callback;
      arg = m->timers[0]->arg;
      remove_timer(m, m->timers[0]);
      callback(arg, MUX_SUCCESS);
    }

  /* Process input and output events. */
  for (mfd = m->inputs; mfd; mfd = next)
    {
      next = mfd->next;
      if (FD_ISSET(mfd->fd, read_fds))
	mfd->callback(mfd->arg, MUX_SUCCESS);
    }
  for (mfd = m->outputs; mfd; mfd = next)
    {
      next = mfd->next;
      if (FD_ISSET(mfd->fd, write_fds))
	mfd->callback(mfd->arg, MUX_SUCCESS);
    }
}

static void remove_timer(mux m, mux_timer timer)
{
  int pos, ppos, lowest;

  /* Free the timer, saving its heap position and argument. */
  pos = timer->heap_pos;
  free(timer);

  if (pos != m->num_timers - 1)
    {
      /* Replace the timer with the last timer in the heap and
       * restore the heap, propagating the timer either up or
       * down, depending on which way it violates the heap
       * property to insert the last timer in place of the
       * deleted timer.
       */
      ppos = parent(pos);
      if (pos > 0 && hcomp(m->timers, m->num_timers - 1, ppos) < 0)
	{
	  do
	    {
	      hassign(m->timers, pos, m->timers[ppos]);
	      pos = ppos;
	      ppos = parent(pos);
	    }
	  while (pos > 0 && hcomp(m->timers, m->num_timers - 1, ppos) < 0);
	  hassign(m->timers, pos, m->timers[m->num_timers - 1]);
	}
      else
	{
	  while (child2(pos) < m->num_timers)
	    {
	      lowest = m->num_timers - 1;
	      if (hcomp(m->timers, child1(pos), lowest) < 0)
		lowest = child1(pos);
	      if (hcomp(m->timers, child2(pos), lowest) < 0)
		lowest = child2(pos);
	      hassign(m->timers, pos, m->timers[lowest]);
	      pos = lowest;
	    }
	  if (pos != m->num_timers - 1)
	    hassign(m->timers, pos, m->timers[m->num_timers - 1]);
	}
    }
  m->num_timers--;
}

static int parent(int n)
{
  return (n - 1) / 2;
}

static int child1(int n)
{
  return n * 2 + 1;
}

static int child2(int n)
{
  return n * 2 + 2;
}

static void hassign(mux_timer *heap, int pos, mux_timer timer)
{
  heap[pos] = timer;
  timer->heap_pos = pos;
}

static int tcomp(mux_timer t1, mux_timer t2)
{
  return tvcomp(&t1->abstime, &t2->abstime);
}

static int hcomp(mux_timer *heap, int pos1, int pos2)
{
  return tvcomp(&heap[pos1]->abstime, &heap[pos2]->abstime);
}

static int tvcomp(struct timeval *tv1, struct timeval *tv2)
{
  if (tv1->tv_sec < tv2->tv_sec)
    return -1;
  else if (tv1->tv_sec > tv2->tv_sec)
    return 1;
  else if (tv1->tv_usec < tv2->tv_usec)
    return -1;
  else if (tv1->tv_usec > tv2->tv_usec)
    return 1;
  else
    return 0;
}

static void tvsub(struct timeval *result, struct timeval *tv1,
		  struct timeval *tv2)
{
  if (tv1->tv_sec < tv2->tv_sec || (tv1->tv_sec == tv2->tv_sec
				    && tv1->tv_usec < tv2->tv_usec))
    {
      result->tv_sec = 0;
      result->tv_usec = 0;
    }
  else
    {
      result->tv_sec = tv1->tv_sec - tv2->tv_sec;
      if (tv1->tv_usec < tv2->tv_usec)
	{
	  result->tv_sec--;
	  result->tv_usec = 1000000 - (tv2->tv_usec - tv1->tv_usec);
	}
      else
	result->tv_usec = tv1->tv_usec - tv2->tv_usec;
    }
}
