/*
 * port.c --- the Port abstraction for communicating with the modem
 *
 * Written by Theodore Ts'o, 1999
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <termios.h>
#include <ctype.h>

#include "loadrm2.h"
#include "rocket.h"

#define STATE_SEND_ATI		1
#define STATE_SEND_START_DNLD	2
#define STATE_LOADER_ACK	3
#define STATE_FIRMWARE_READY	4
#define STATE_FIRMWARE		5
#define STATE_CHECKSUM		6
#define STATE_DONE		7

char *send_str_ati = "ATI3\r";
char *send_str_start_dnld = "AT**\r";

char *recv_str_OK = "ok\r";
char *recv_str_loaded = "v2.101a2-v90_2m_dls_ram";
char *recv_str_start_dnld = "download initiated";
char *recv_str_start_frmwr = "at 115.2k";
char *recv_str_programmed = "device successfully programmed";

static int bytes_written = 0;
static int max_bytes_written = 0;

int num_ports = 0, ports_ready = 0;
	
struct port {
    char	*name;
    int		fd;
    int		state;
    mux		m;
    mux_fd	read_mux;
    mux_fd	write_mux;
    mux_timer	timeout_mux;
    int		eof_count;
    int		code_ptr;
    char	read_buf[READBUFSIZE];
    char	*read_ptr;
//    const char	*write_ptr;
    unsigned char	*write_ptr;
    int		write_len;
    time_t	start_time;
    int		retries;
    int		tried_reset;
    struct port	*next;
};

struct port *port_list = 0;

static void port_timeout_callback(void *arg, int status);

/*
 * This shuts down a port
 */
void port_shutdown(struct port *port)
{
    if (port->read_mux) {
	mux_cancel_fd(port->m, port->read_mux);
	port->read_mux = 0;
    }
    if (port->write_mux) {
	mux_cancel_fd(port->m, port->write_mux);
	port->write_mux = 0;
    }
    if (port->timeout_mux) {
	mux_cancel_timer(port->m, port->timeout_mux);
	port->timeout_mux = 0;
    }
}

//  This callback function handles all of the generic writing to the port
//  Called by the mux when a write will not block on the port, select is used().
//  The data to write is stored in the port struct, passed as a void *.
void port_write_callback(void *arg, int status)
{
    struct port *port = arg;
    int		len;
	
    if (status == MUX_EDESTRUCTION)
	return;		/* No cleanup necessary */

    if (port->write_len && port->write_ptr) {
	len = write(port->fd, port->write_ptr, port->write_len);
	if (len < 0) {
	    if (errno == EINTR || errno == EAGAIN)
		return;
	    /* some kind of error handling */
	    return;
	}
	if (debug_flag) {
	    printf("%s: %d bytes written\n", port->name, len);
	    fflush(stdout);
	}
	if (port->state == STATE_FIRMWARE) {
	    bytes_written += len;
	    if (progress_flag)
		progress_update();
	}
	port->write_len -= len;
	port->write_ptr += len;
    }
    if (port->write_len <= 0) {
	mux_cancel_fd(port->m, port->write_mux);
	port->write_mux = 0;
	port->write_len = 0;
	port->write_ptr = 0;
    }
}


//  This is how we send short commands to the modem.
//  Adds a string to the write buffer, starts the write callback.
//  The mux does a select() on the fd, when the write will not block
//  the callback (above) is called.
void port_send_string(struct port *port, const char *str)
{
    if (debug_flag)
	printf("%s: Sending string to modem: %s\n", port->name, str);
//    port->write_ptr = str;
    port->write_ptr = (unsigned char *)str;
    port->write_len = strlen(str);
    if (port->write_mux)
	mux_cancel_fd(port->m, port->write_mux);
    mux_add_output(port->m, &port->write_mux, port->fd,
		   port_write_callback, port);
}

/*
 * And how we download code to the modem...
 */
void port_send_code(struct port *port, code_t code)
{
    if (debug_flag)
	printf("%s: Sending code to modem...\n", port->name);
    port->write_ptr = code->data;	//both are now unsigned char pointers in v3.04 - krw
    port->write_len = code->length;
    if (port->write_mux)
	mux_cancel_fd(port->m, port->write_mux);
    mux_add_output(port->m, &port->write_mux, port->fd,
		   port_write_callback, port);
}

/*
 * This function sends the appropriate string to the port based on the
 * current state of the FSM
 */
void port_send_cmd(struct port *port)
{
    if (port->state == STATE_SEND_ATI) {
	port_send_string(port, send_str_ati);
	if (port->timeout_mux)
	    mux_cancel_timer(port->m, port->timeout_mux);
	mux_add_rstimer(port->m, &port->timeout_mux, 1,
			port_timeout_callback, port);
    } 

    //  Sends the start download string to the port, starts the timeout timer.
    else if (port->state == STATE_SEND_START_DNLD) {
	port_send_string(port, send_str_start_dnld);
	mux_add_rstimer(port->m, &port->timeout_mux, 5,
			port_timeout_callback, port);
    } 

    //  Starts sending the firmware to the port
    else if (port->state == STATE_LOADER_ACK) {
	port_send_code(port, loader_code);
    }
}

/*
 * This function causes us to kick off the firmware download, after
 * waiting a brief delay to avoid overrunning the modems
 */
static void start_firmware_callback(void *arg, int status);

static void start_firmware(mux m)
{
    struct timeval 	tv;
			
    gettimeofday(&tv, 0);
    tv.tv_usec += 500000; /* 0.5 seconds */
    mux_add_timer(m, 0, &tv, start_firmware_callback, m);
}

/*
 * This is the callback function for start_firmware()
 */
static void start_firmware_callback(void *arg, int status)
{
    struct port *port;

    if (status == MUX_EDESTRUCTION)
	return;
    progress_init();

    for (port = port_list; port; port = port->next) {
		if (port->state == STATE_FIRMWARE_READY) {
		    port_send_code(port, firmware_code);
		    port->state = STATE_FIRMWARE;
		}
    }
}


/*
 * Callback function to force a reset of a modem.  After we reset,
 * retry the current state command...
 */
#define PORT_RESET_OK		0
#define PORT_RESET_NOTOK	1

static void port_reset_callback1(void *arg, int status);
static void port_reset_callback2(void *arg, int status);

//  Send the port reset IOCTL to the port, start a timeout timer.
static int port_reset(struct port *port)
{
    int		ret;
    int		reset_arg = 1;

    ret = ioctl(port->fd, RCKP_RESET_RM2, &reset_arg);
    if (ret < 0) 
	return PORT_RESET_NOTOK;

    mux_add_rstimer(port->m, &port->timeout_mux, 1,
		    port_reset_callback1, port);

    return PORT_RESET_OK;
}

/*
 * Callback function #1 for port_reset()
 */
static void port_reset_callback1(void *arg, int status)
{
    struct port *port = arg;
    int	reset_arg = 0;
			
    port->timeout_mux = 0;	/* timers are one-shot */
    if (status == MUX_EDESTRUCTION)
	return;		/* No cleanup necessary */

    ioctl(port->fd, RCKP_RESET_RM2, &reset_arg);
    mux_add_rstimer(port->m, &port->timeout_mux, 1,
		    port_reset_callback2, port);
}

/*
 * Callback function #2 for port_reset()
 */
static void port_reset_callback2(void *arg, int status)
{
    struct port *port = arg;

    port->timeout_mux = 0;	/* timers are one-shot */
    if (status == MUX_EDESTRUCTION)
	return;		/* No cleanup necessary */

    port_send_cmd(port);	/* Retry the current command */
}


//  This is our timeout function, in case a port doesn't respond.
//  This depends on the state.  During STATE_SEND_START_DNLD it initiates
//  a single reset attempt.  The time is also active during successful 
//  firmware load, it expires silently here since it falls through the 
//  switch statement.  Note that this does open up the possibility of a
//  successful start to firmware load, but a timeout occuring that would
//  hang and wait forever.
static void port_timeout_callback(void *arg, int status)
{
    struct port *port = arg;

    port->timeout_mux = 0;	/* timers are one-shot */

    if (status == MUX_EDESTRUCTION)
	return;		/* No cleanup necessary */

    switch (port->state) {
    case STATE_SEND_ATI:
	if (port->retries++ < 3) {
	    port_send_cmd(port);
	    break;
	}
	/* fall through */
    case STATE_SEND_START_DNLD:
	printf("No response from RocketModem on %s, ", port->name);
	if (!port->tried_reset) {
	    port->tried_reset = 1;
	    port->retries = 0;
	    printf("trying reset...\n");
	    if (port_reset(port) == PORT_RESET_OK) 
		return;
	    printf("Reset of %s failed!\n", port->name);
	} else
	    printf("couldn't initialize.\n");
	port->state = STATE_DONE;
	port_shutdown(port);
	num_ports--;
	if (ports_ready >= num_ports) 
	    start_firmware(port->m);
    }
}

static void port_timeout_cancel(struct port *port)
{
    if (port->timeout_mux) {
	mux_cancel_timer(port->m, port->timeout_mux);
	port->timeout_mux = 0;
    }
}		


//  This callback function handles all of the generic reading from the port
//  Called by the mux when a read will not block on the port, select is used().
//  The port struct, passed as a void *, contain info on what port to read
//  and storage for the data.
void port_read_callback(void *arg, int status)
{
    struct port *port = arg;
    int		maxlen, i, len;
    char		*cp;
	
    if (status == MUX_EDESTRUCTION)
	return;		/* No cleanup necessary */

    maxlen = READBUFSIZE - (port->read_ptr - port->read_buf) - 1;

    len = read(port->fd, port->read_ptr, maxlen);
    if (len < 0) {
	if ((errno == EINTR) || (errno == EAGAIN))
	    return;
	len = 0;
    }
    for (i=0; i < len; i++) {
	if (port->read_ptr[i] == '\0')
	    port->read_ptr[i] = ' ';
	if (isupper(port->read_ptr[i]))
	    port->read_ptr[i] = tolower(port->read_ptr[i]);
    }
    if (len) {
	port->read_ptr += len;
	port->eof_count = 0;
    } else {
	port->eof_count++;
	/* XXX do something if EOF gets too high */
    }

    *(port->read_ptr) = '\0';

    switch (port->state) {
    case STATE_SEND_ATI:
	if (strstr(port->read_buf, recv_str_loaded)) {
	    if (debug_flag)
		printf("%s already loaded.\n", port->name);
	    port->state = STATE_DONE;
	    port_shutdown(port);
	    num_ports--;
	} else if (strstr(port->read_buf, recv_str_OK)) {
	    port->state = STATE_SEND_START_DNLD;
	    port_send_cmd(port);
	}
	break;

    //  Receives the response to the AT** command send to the port.
    //  If OK, advances the state and sends the next command.
    case STATE_SEND_START_DNLD:
	if (strstr(port->read_buf, recv_str_start_dnld)) {
	    port->state = STATE_LOADER_ACK;
	    port_send_cmd(port);
	}
	break;

    // Receives the port's response to a firmware load.  If load is done,
    // advances the state and starts the firmware. 
    case STATE_LOADER_ACK:
	if (strstr(port->read_buf, recv_str_start_frmwr)) {
	    port->state = STATE_FIRMWARE_READY;
	    port_timeout_cancel(port);
	    ports_ready++;
	    if (ports_ready >= num_ports)
		start_firmware(port->m);
	}
	break;

    // Firmware is ready, close the port
    case STATE_FIRMWARE:
	if (strstr(port->read_buf, recv_str_programmed)) {
	    port->state = STATE_DONE;
	    port_shutdown(port);
	}
	break;
    }

    while ((cp = index(port->read_buf, '\r')) != 0) {
	if (*cp)
	    *cp = 0;
	else
	    cp[1] = 0;
	if (debug_flag == 2)
	    printf("Port %s read: '%s'\n", port->name,
		   port->read_buf);
	memmove(port->read_buf, cp+1, strlen(cp+1)+1);
	port->read_ptr -= (cp+1) - port->read_buf;
    }

}

int initialize_termios(int fd)
{
    struct termios	t;

    if (tcgetattr(fd, &t) < 0)
	return errno;

    cfsetospeed(&t, B115200);
    cfsetispeed(&t, B115200);
    t.c_iflag &= ~(ISTRIP|INLCR|ICRNL);
    t.c_oflag &= ~(OPOST|ONLCR|OCRNL|ONLRET|OFILL);
    t.c_cflag |= CREAD|CLOCAL;
    t.c_lflag &= ~(ISIG|ICANON|ECHO|IEXTEN);

    if (tcsetattr(fd, TCSANOW, &t) < 0)
	return errno;

    return 0;
}


//  Opens a port (presumable a RocketModem II).  Creates a port struct,
//  fills it in.  Registers a mux read callback, sets state to start download.
//  Sends download command to port, which also starts a timer (and registers 
//  a timeout callback).  
//  
//  To start download to the flash, you send AT**.  The port responds with 
//  "download initiated".  You then send flash code to it.
//
//  The state maching starts in STATE_SEND_START_DNLD.
//  This function sends the AT**, the mux read object waits for a response.
//  A 5 second timer is started for timeout purposes.  If the AT** response
//  is recieved (see port_read_callback()) before the timeout, the timer is 
//  canceled.  The state is advanced to STATE_LOADER_ACK and the load begins.
int port_open(mux m, char *name)
{
    struct port *port;
    int	fd;

    if ((fd = open(name, O_RDWR|O_NONBLOCK)) < 0)
	return errno;

    port = malloc(sizeof(struct port));
    if (!port)
	return ENOMEM;
    memset(port, 0, sizeof(struct port));

    port->name = malloc(strlen(name)+1);
    if (!port->name) {
	free(port);
	return ENOMEM;
    }
    strcpy(port->name, name);
    initialize_termios(fd);
    port->fd = fd;
    port->read_ptr = port->read_buf;
    port->start_time = time(0);
    port->m = m;
    port->next = port_list;
    port_list = port;

    //  Add the read callback to the mux
    mux_add_input(m, &port->read_mux, fd, port_read_callback, port);

    // Set state to download flash code
    port->state = ati_flag ? STATE_SEND_ATI : STATE_SEND_START_DNLD;
    if (ati_flag == 2)
	port->state = STATE_LOADER_ACK;
    port->write_mux = 0;
    port->timeout_mux = 0;

    //  Starts the download by sending a string to the port.
    //  This also starts the timeout timer.
    port_send_cmd(port);
    num_ports++;
    return 0;
}


/*
 * Here are the progress reporting functions....
 */
#define PROGRESS_BAR_SIZE 60

static char bar_chars[PROGRESS_BAR_SIZE+1];
static char empty_chars[PROGRESS_BAR_SIZE+1];
static char blank_chars[PROGRESS_BAR_SIZE+1];
static char backup_chars[PROGRESS_BAR_SIZE+1];

static int previous_progress;

void progress_init()
{
    bytes_written = 0;
    max_bytes_written = num_ports * firmware_code->length;
    previous_progress = 0;
    if (!progress_flag)
	return;
    printf("Loading firmware: ");
    memset(bar_chars, 	'=', PROGRESS_BAR_SIZE);
    memset(empty_chars, 	'.', PROGRESS_BAR_SIZE);
    memset(blank_chars, 	' ', PROGRESS_BAR_SIZE);
    memset(backup_chars, 	'\b', PROGRESS_BAR_SIZE);
    bar_chars[PROGRESS_BAR_SIZE] = 0;
    empty_chars[PROGRESS_BAR_SIZE] = 0;
    blank_chars[PROGRESS_BAR_SIZE] = 0;
    backup_chars[PROGRESS_BAR_SIZE] = 0;
    fputs(empty_chars, stdout);
    fputs(backup_chars, stdout);
    fflush(stdout);
}

void progress_update()
{
    int	prog;

    if (!progress_flag)
	return;
    prog = PROGRESS_BAR_SIZE*bytes_written/max_bytes_written;
    if (prog == previous_progress)
	return;

    fputs(bar_chars+(PROGRESS_BAR_SIZE-prog), stdout);
    fputs(backup_chars+(PROGRESS_BAR_SIZE-prog), stdout);
    fflush(stdout);
    previous_progress = prog;
}

void progress_end()
{
    if (!progress_flag)
	return;
    fputs(blank_chars, stdout);
    fputs(backup_chars, stdout);
    fputs("done\n", stdout);
}
