/*
 * RocketPort device driver for Linux
 *
 * Written by Theodore Ts'o, 1995, 1996, 1997, 1998, 1999, 2000.
 *
 * Copyright (C) 1995, 1996, 1997, 1998, 1999, 2000, 2003 by Comtrol, Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * Kernel Synchronization:
 *
 * This driver has 2 kernel control paths - exception handlers (calls into the driver
 * from user mode) and the timer bottom half (tasklet).  This is a polled driver, interrupts
 * are not used.
 *
 * Critical data:
 * -  rp_table[], accessed through passed "info" pointers, is a global (static) array of
 *    serial port state information and the xmit_buf circular buffer.  Protected by
 *    a per port spinlock.
 * -  xmit_flags[], an array of ints indexed by line (port) number, indicating that there
 *    is data to be transmitted.  Protected by atomic bit operations.
 * -  rp_num_ports, int indicating number of open ports, protected by atomic operations.
 *
 */

/*******  Driver version number and date *******/
#define ROCKET_VERSION "3.16"
#define ROCKET_DATE "11-Aug-2011"

/****** Defines ******/

#ifndef LINUX_VERSION_CODE
#  include <linux/version.h>
#endif

#ifdef PCI_NUM_RESOURCES
#define PCI_BASE_ADDRESS(dev, r) ((dev)->resource[r].start)
#else
#define PCI_BASE_ADDRESS(dev, r) ((dev)->base_address[r])
#endif

#ifndef VERSION_CODE
#  define VERSION_CODE(vers,rel,seq) ( ((vers)<<16) | ((rel)<<8) | (seq) )
#endif

#if LINUX_VERSION_CODE < VERSION_CODE(2,2,9)	/*  No version < 2.2 */
#  error "This kernel is too old: not supported by this file"
#endif

#ifdef MODVERSIONS

#if LINUX_VERSION_CODE < VERSION_CODE(2,5,00)
#include <linux/modversions.h>
#else
#include <config/modversions.h>
#endif

#endif

/*
* users of non-2.6.16 kernels will have to modify the version supplied to
* allow complilation of new tty layer handing code.
* krw - 2006.04.08
*/
#if !defined(TTYLAYER_NEW) && (LINUX_VERSION_CODE >= VERSION_CODE(2,6,15))
#define TTYLAYER_NEW
#endif

#ifdef TTYLAYER_NEW
#define ReceiveRoom(tty)  tty->receive_room
#else
#define ReceiveRoom(tty)  tty->ldisc.receive_room(tty)
#endif


#define ROCKET_PARANOIA_CHECK
#define ROCKET_DISABLE_SIMUSAGE

#undef ROCKET_SOFT_FLOW
#undef ROCKET_DEBUG_OPEN
#undef ROCKET_DEBUG_INTR
#undef ROCKET_DEBUG_WRITE
#undef ROCKET_DEBUG_FLOW
#undef ROCKET_DEBUG_THROTTLE
#undef ROCKET_DEBUG_WAIT_UNTIL_SENT
#undef ROCKET_DEBUG_RECEIVE
#undef ROCKET_DEBUG_HANGUP
#undef REV_PCI_ORDER
#undef ROCKET_DEBUG_IO
#undef ROCKET_DEBUG_DELTA_MSR

#undef CONFIG_DEVFS_FS

#if LINUX_VERSION_CODE < VERSION_CODE(2,4,0)
#define TTY_DRIVER_NO_DEVFS 0
#endif

#define POLL_PERIOD HZ/100	/*  Polling period .01 seconds (10ms) */

/****** Kernel includes ******/

#ifdef MODVERSIONS
#if LINUX_VERSION_CODE < VERSION_CODE(2,5,00)
#include <linux/modversions.h>
#else
#include <config/modversions.h>
#endif
#endif

#include <linux/module.h>
#include <linux/errno.h>
#include <linux/major.h>
#include <linux/kernel.h>
#include <linux/signal.h>
#include <linux/wait.h>

#if LINUX_VERSION_CODE < VERSION_CODE(2,4,0)
#include <linux/malloc.h>
#else
#include <linux/slab.h>
#endif

#include <linux/mm.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/interrupt.h>
#include <linux/tty.h>
#include <linux/tty_driver.h>
#include <linux/tty_flip.h>
#include <linux/string.h>
#include <linux/fcntl.h>
#include <linux/ptrace.h>
#include <linux/major.h>
#include <linux/ioport.h>
#include <linux/delay.h>
#include <linux/wait.h>
#include <linux/delay.h>
#include <linux/pci.h>
#include <asm/uaccess.h>
#include <asm/atomic.h>
#include <asm/bitops.h>
#if LINUX_VERSION_CODE < VERSION_CODE(2,4,0)
#include <asm/spinlock.h>
#else
#include <linux/spinlock.h>
#endif
#if LINUX_VERSION_CODE < VERSION_CODE(2,6,27)
#include <asm/semaphore.h>
#else
#include <linux/semaphore.h>
#endif
#include <linux/init.h>

/****** RocketPort includes ******/

#include "rocket_int.h"
#include "rocket.h"
//the following line is for the stripped kernel version - krw 2K6.19.06
//#include "rocket_config.h"

//make sure this is done after tty.h is included.
#ifndef TTY_FLIPBUF_SIZE
#define TTY_FLIPBUF_SIZE 512
#endif

/****** RocketPort Local Variables ******/
#if LINUX_VERSION_CODE < VERSION_CODE(2,5,0)
static struct tty_struct *rocket_table[MAX_PORTS];	/*  TTY required variables */
static int rocket_refcount;
static struct termios *rocket_termios[MAX_PORTS];
static struct termios *rocket_termios_locked[MAX_PORTS];
#endif

static struct tty_driver *rocket_driver;

static struct rocket_version driver_version = { ROCKET_VERSION, ROCKET_DATE };

static struct r_port *rp_table[MAX_PORTS];	/*  The main repository of serial port state information. */
static unsigned int xmit_flags[MAX_NUMBEROF_BOARDS];	/*  Bit significant, indicates port had data to transmit. */
/*  eg.  Bit 0 indicates port 0 has xmit data, ...        */
static atomic_t rp_num_ports_open;	/*  Number of serial ports open                           */
static int is_PCI[MAX_NUMBEROF_BOARDS];	/*  Flag needed by cleanup_module, so we release only ISA IO regions, not PCI */
static struct timer_list rocket_timer;

static int numberof_boards;	/*  The number of boards installed (ISA + PCI) */

static unsigned long board1;	/* ISA addresses, retrieved from rocketport.conf */
static unsigned long board2;
static unsigned long board3;
static unsigned long board4;
static unsigned long controller;
static unsigned long support_low_speed;
static unsigned long modem1;
static unsigned long modem2;
static unsigned long modem3;
static unsigned long modem4;
static unsigned long pc104_1[8];
static unsigned long pc104_2[8];
static unsigned long pc104_3[8];
static unsigned long pc104_4[8];
static unsigned long *pc104[4] = { pc104_1, pc104_2, pc104_3, pc104_4 };
static unsigned long isa_io_addr[MAX_NUMBEROF_BOARDS];	/*  ISA IO base address */
static int isa_type[MAX_NUMBEROF_BOARDS];	/*  ISA board type */

static CONTROLLER_T sController[MAX_NUMBEROF_BOARDS];	/*  Global array of configuration data, indexed by board number */

/*  Receive Processor Microcode */
static const u8 RData[RDATASIZE] = {
	0x00, 0x09, 0xf6, 0x82,
	0x02, 0x09, 0x86, 0xfb,
	0x04, 0x09, 0x00, 0x0a,
	0x06, 0x09, 0x01, 0x0a,
	0x08, 0x09, 0x8a, 0x13,
	0x0a, 0x09, 0xc5, 0x11,
	0x0c, 0x09, 0x86, 0x85,
	0x0e, 0x09, 0x20, 0x0a,
	0x10, 0x09, 0x21, 0x0a,
	0x12, 0x09, 0x41, 0xff,
	0x14, 0x09, 0x82, 0x00,
	0x16, 0x09, 0x82, 0x7b,
	0x18, 0x09, 0x8a, 0x7d,
	0x1a, 0x09, 0x88, 0x81,
	0x1c, 0x09, 0x86, 0x7a,
	0x1e, 0x09, 0x84, 0x81,
	0x20, 0x09, 0x82, 0x7c,
	0x22, 0x09, 0x0a, 0x0a
};

static const u8 RRegData[RREGDATASIZE] = {
	0x00, 0x09, 0xf6, 0x82,	/* 00: Stop Rx processor */
	0x08, 0x09, 0x8a, 0x13,	/* 04: Tx software flow control */
	0x0a, 0x09, 0xc5, 0x11,	/* 08: XON char */
	0x0c, 0x09, 0x86, 0x85,	/* 0c: XANY */
	0x12, 0x09, 0x41, 0xff,	/* 10: Rx mask char */
	0x14, 0x09, 0x82, 0x00,	/* 14: Compare/Ignore #0 */
	0x16, 0x09, 0x82, 0x7b,	/* 18: Compare #1 */
	0x18, 0x09, 0x8a, 0x7d,	/* 1c: Compare #2 */
	0x1a, 0x09, 0x88, 0x81,	/* 20: Interrupt #1 */
	0x1c, 0x09, 0x86, 0x7a,	/* 24: Ignore/Replace #1 */
	0x1e, 0x09, 0x84, 0x81,	/* 28: Interrupt #2 */
	0x20, 0x09, 0x82, 0x7c,	/* 2c: Ignore/Replace #2 */
	0x22, 0x09, 0x0a, 0x0a	/* 30: Rx FIFO Enable */
};

static const u8 sBitMapClrTbl[8] = {
	0xfe, 0xfd, 0xfb, 0xf7, 0xef, 0xdf, 0xbf, 0x7f
};

static const u8 sBitMapSetTbl[8] = {
	0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80
};

/*
 * The following arrays define the interrupt bits corresponding to each AIOP.
 * These bits are different between the ISA/PCI boards and the UPCI boards.
 */
static const u32 aiop_intr_bits[MAX_AIOPS_PER_BOARD] = {
	AIOP_INTR_BIT_0,
	AIOP_INTR_BIT_1,
	AIOP_INTR_BIT_2,
	AIOP_INTR_BIT_3
};

static const u32 upci_aiop_intr_bits[MAX_AIOPS_PER_BOARD] = {
	UPCI_AIOP_INTR_BIT_0,
	UPCI_AIOP_INTR_BIT_1,
	UPCI_AIOP_INTR_BIT_2,
	UPCI_AIOP_INTR_BIT_3
};

/*
 *  Line number is the ttySIx number (x), the Minor number.  We
 *  assign them sequentially, starting at zero.  The following
 *  array keeps track of the line number assigned to a given board/aiop/channel.
 */
static unsigned char lineNumbers[MAX_PORTS];
static unsigned long nextLineNumber;
static int totalPorts = 0;

/*****  RocketPort Static Prototypes   *********/
static int __init init_ISA(int i);
static void rp_wait_until_sent(struct tty_struct *tty, int timeout);
static void rp_flush_buffer(struct tty_struct *tty);
static unsigned char GetLineNumber(int ctrl, int aiop, int ch);
static unsigned char SetLineNumber(int ctrl, int aiop, int ch);
static void rp_start(struct tty_struct *tty);
#if LINUX_VERSION_CODE < VERSION_CODE(2,5,0)
static void check_for_delta_msr(struct r_port *info);
#endif
static int sInitController(CONTROLLER_T * CtlP, int CtlNum, u32 MudbacIO,
			   u32 * AiopIOList, int AiopIOListSize,
			   int IRQNum, u8 Frequency, int PeriodicOnly);

static int sReadAiopID(u32 io);
static int sReadAiopNumChan(u32 io);

static int sInitChan(CONTROLLER_T * CtlP, CHANNEL_T * ChP, int AiopNum, int ChanNum);
static void sStopRxProcessor(CHANNEL_T * ChP);
static void sFlushRxFIFO(CHANNEL_T * ChP);
static void sFlushTxFIFO(CHANNEL_T * ChP);
static int sWriteTxPrioByte(CHANNEL_T * ChP, u8 Data);
static void sEnInterrupts(CHANNEL_T * ChP, u32 Flags);
static void sDisInterrupts(CHANNEL_T * ChP, u32 Flags);
static void sSetInterfaceMode(CHANNEL_T * ChP, u8 mode);
static void sModemReset(CONTROLLER_T * CtlP, int chan, int on);
static void sPCIModemReset(CONTROLLER_T * CtlP, int chan, int on);
static void rmSpeakerReset(CONTROLLER_T * CtlP, int model);
static int GetNumberOfModels(void);
static modelInfo_t *GetModelInfo(int model);
static void SetClockPrescale(CONTROLLER_T * CtlP, int aiopnum, unsigned char prescale);

#ifdef MODULE
MODULE_AUTHOR("Theodore Ts'o");
MODULE_DESCRIPTION("Comtrol RocketPort driver");
//MODULE_PARM is history in vanilla 2.6.17 but been around since 2.6.0
//per http://lwn.net/Articles/22028
#if LINUX_VERSION_CODE < VERSION_CODE(2,6,00)
MODULE_PARM(board1, "i");
MODULE_PARM_DESC(board1, "I/O port for (ISA) board #1");
MODULE_PARM(board2, "i");
MODULE_PARM_DESC(board2, "I/O port for (ISA) board #2");
MODULE_PARM(board3, "i");
MODULE_PARM_DESC(board3, "I/O port for (ISA) board #3");
MODULE_PARM(board4, "i");
MODULE_PARM_DESC(board4, "I/O port for (ISA) board #4");
MODULE_PARM(controller, "i");
MODULE_PARM_DESC(controller, "I/O port for (ISA) rocketport controller");
MODULE_PARM(support_low_speed, "i");
MODULE_PARM_DESC(support_low_speed,"1 means support 50 baud, 0 means support 460400 baud");
MODULE_PARM(modem1, "i");
MODULE_PARM_DESC(modem1, "1 means (ISA) board #1 is a RocketModem");
MODULE_PARM(modem2, "i");
MODULE_PARM_DESC(modem2, "1 means (ISA) board #2 is a RocketModem");
MODULE_PARM(modem3, "i");
MODULE_PARM_DESC(modem3, "1 means (ISA) board #3 is a RocketModem");
MODULE_PARM(modem4, "i");
MODULE_PARM_DESC(modem4, "1 means (ISA) board #4 is a RocketModem");
MODULE_PARM(pc104_1, "1-8i");
MODULE_PARM_DESC(pc104_1,"set interface types for ISA(PC104) board #1 (e.g. pc104_1=232,232,485,485,...");
MODULE_PARM(pc104_2, "1-8i");
MODULE_PARM_DESC(pc104_2,"set interface types for ISA(PC104) board #2 (e.g. pc104_2=232,232,485,485,...");
MODULE_PARM(pc104_3, "1-8i");
MODULE_PARM_DESC(pc104_3,"set interface types for ISA(PC104) board #3 (e.g. pc104_3=232,232,485,485,...");
MODULE_PARM(pc104_4, "1-8i");
MODULE_PARM_DESC(pc104_4,"set interface types for ISA(PC104) board #4 (e.g. pc104_4=232,232,485,485,...");
#else
module_param(board1, ulong, 0);
MODULE_PARM_DESC(board1, "I/O port for (ISA) board #1");
module_param(board2, ulong, 0);
MODULE_PARM_DESC(board2, "I/O port for (ISA) board #2");
module_param(board3, ulong, 0);
MODULE_PARM_DESC(board3, "I/O port for (ISA) board #3");
module_param(board4, ulong, 0);
MODULE_PARM_DESC(board4, "I/O port for (ISA) board #4");
module_param(controller, ulong, 0);
MODULE_PARM_DESC(controller, "I/O port for (ISA) rocketport controller");
module_param(support_low_speed, ulong, 0);
MODULE_PARM_DESC(support_low_speed, "1 means support 50 baud, 0 means support 460400 baud");
module_param(modem1, ulong, 0);
MODULE_PARM_DESC(modem1, "1 means (ISA) board #1 is a RocketModem");
module_param(modem2, ulong, 0);
MODULE_PARM_DESC(modem2, "1 means (ISA) board #2 is a RocketModem");
module_param(modem3, ulong, 0);
MODULE_PARM_DESC(modem3, "1 means (ISA) board #3 is a RocketModem");
module_param(modem4, ulong, 0);
MODULE_PARM_DESC(modem4, "1 means (ISA) board #4 is a RocketModem");
#if LINUX_VERSION_CODE < VERSION_CODE(2,6,10)
int nump = 0;	//needed for module_param_array on 2.6.9 kernels I found.
module_param_array(pc104_1, ulong, nump, 0);
MODULE_PARM_DESC(pc104_1, "set interface types for ISA(PC104) board #1 (e.g. pc104_1=232,232,485,485,...");
module_param_array(pc104_2, ulong, nump, 0);
MODULE_PARM_DESC(pc104_2, "set interface types for ISA(PC104) board #2 (e.g. pc104_2=232,232,485,485,...");
module_param_array(pc104_3, ulong, nump, 0);
MODULE_PARM_DESC(pc104_3, "set interface types for ISA(PC104) board #3 (e.g. pc104_3=232,232,485,485,...");
module_param_array(pc104_4, ulong, nump, 0);
MODULE_PARM_DESC(pc104_4, "set interface types for ISA(PC104) board #4 (e.g. pc104_4=232,232,485,485,...");
#else
module_param_array(pc104_1, ulong, NULL, 0);
MODULE_PARM_DESC(pc104_1, "set interface types for ISA(PC104) board #1 (e.g. pc104_1=232,232,485,485,...");
module_param_array(pc104_2, ulong, NULL, 0);
MODULE_PARM_DESC(pc104_2, "set interface types for ISA(PC104) board #2 (e.g. pc104_2=232,232,485,485,...");
module_param_array(pc104_3, ulong, NULL, 0);
MODULE_PARM_DESC(pc104_3, "set interface types for ISA(PC104) board #3 (e.g. pc104_3=232,232,485,485,...");
module_param_array(pc104_4, ulong, NULL, 0);
MODULE_PARM_DESC(pc104_4, "set interface types for ISA(PC104) board #4 (e.g. pc104_4=232,232,485,485,...");
#endif
#endif
#endif

/*
 *  The Linux convention in kernel 2.6 is to not use init_module() and cleanup_module()
 *  but instead to use driver specific names.
 */
#if LINUX_VERSION_CODE > VERSION_CODE(2,4,99)
static int rp_init(void);
static void rp_cleanup_module(void);

module_init(rp_init);
module_exit(rp_cleanup_module);
#endif

#ifdef MODULE_LICENSE
MODULE_LICENSE("Dual BSD/GPL");
#endif

#define checkpoint() printk(KERN_INFO "RocketPort at line %d\n", __LINE__)

/*************************************************************************/
/*                     Module code starts here                           */

static inline int rocket_paranoia_check(struct r_port *info,
					const char *routine)
{
#ifdef ROCKET_PARANOIA_CHECK
	if (!info)
		return 1;
	if (info->magic != RPORT_MAGIC) {
		printk(KERN_INFO
		       "Warning: bad magic number for RocketPort struct in %s\n",
		       routine);
		return 1;
	}
#endif
	return 0;
}

/*  Serial port receive data function.  Called (from timer poll) when an AIOPIC signals
 *  that receive data is present on a serial port.  Pulls data from FIFO, moves it into the
 *  tty layer.
 */
static void rp_do_receive(struct r_port *info, struct tty_struct *tty, CHANNEL_t * cp, unsigned int ChanStatus)
{
	unsigned int CharNStat;
	int ToRecv, wRecv, space = 0, count = 0;

#ifdef TTYLAYER_NEW
	unsigned char *cbuf, *chead;
	char *fbuf, *fhead;
	struct tty_ldisc *ld;
	ld = tty_ldisc_ref(tty);
	if(!ld)					//if ld == NULL then don't bother.
		goto done;
#else
	unsigned char *cbuf;
	char *fbuf;
#endif

	ToRecv = sGetRxCnt(cp);

	space = ReceiveRoom(tty);

	if (space > 2 * TTY_FLIPBUF_SIZE)
		space = 2 * TTY_FLIPBUF_SIZE;

#ifndef TTYLAYER_NEW
	cbuf = tty->flip.char_buf;
	fbuf = tty->flip.flag_buf;
#endif

	count = 0;

#ifdef ROCKET_DEBUG_INTR
	printk(KERN_INFO "rp_do_receive(%d, %d)...", ToRecv, space);
#endif

	/*
	 * determine how many we can actually read in.  If we can't
	 * read any in then we have a software overrun condition.
	 */
	if (ToRecv > space)
		ToRecv = space;

#ifdef TTYLAYER_NEW
	ToRecv = tty_prepare_flip_string_flags(tty, &chead, &fhead, ToRecv);
	if (ToRecv <= 0)
		goto done;

	cbuf = chead;
	fbuf = fhead;
#else
	if (ToRecv <= 0)
		return;
#endif

	/*
	 * if status indicates there are errored characters in the
	 * FIFO, then enter status mode (a word in FIFO holds
	 * character and status).
	 */
	if (ChanStatus & (RXFOVERFL | RXBREAK | RXFRAME | RXPARITY)) {
		if (!(ChanStatus & STATMODE)) {
#ifdef ROCKET_DEBUG_RECEIVE
			printk(KERN_INFO "Entering STATMODE...");
#endif

			ChanStatus |= STATMODE;

			sEnRxStatusMode(cp);
		}
	}

	/*
	 * if we previously entered status mode, then read down the
	 * FIFO one word at a time, pulling apart the character and
	 * the status.  Update error counters depending on status
	 */
	if (ChanStatus & STATMODE) {
#ifdef ROCKET_DEBUG_RECEIVE
		printk(KERN_INFO "Ignore %x, read %x...", info->ignore_status_mask, info->read_status_mask);
#endif

		while (ToRecv) {
			CharNStat = sInW(sGetTxRxDataIO(cp));

#ifdef ROCKET_DEBUG_RECEIVE
			printk(KERN_INFO "%x...", CharNStat);
#endif

			if (CharNStat & STMBREAKH)
				CharNStat &= ~(STMFRAMEH | STMPARITYH);
			if (CharNStat & info->ignore_status_mask) {
				ToRecv--;
				continue;
			}
			CharNStat &= info->read_status_mask;
			if (CharNStat & STMBREAKH)
				*fbuf++ = TTY_BREAK;
			else if (CharNStat & STMPARITYH)
				*fbuf++ = TTY_PARITY;
			else if (CharNStat & STMFRAMEH)
				*fbuf++ = TTY_FRAME;
			else if (CharNStat & STMRCVROVRH)
				*fbuf++ = TTY_OVERRUN;
			else
				*fbuf++ = TTY_NORMAL;

			*cbuf++ = CharNStat & 0xff;
			count++;
			ToRecv--;
		}

		/*
		 * after we've emptied the FIFO in status mode, turn
		 * status mode back off
		 */
		if (sGetRxCnt(cp) == 0) {
#ifdef ROCKET_DEBUG_RECEIVE
			printk(KERN_INFO "Status mode off.\n");
#endif
			sDisRxStatusMode(cp);
		}
	} else {
		/*
		 * we aren't in status mode, so read down the FIFO two
		 * characters at time by doing repeated word IO
		 * transfer.
		 */
		wRecv = ToRecv >> 1;
		if (wRecv)
			sInStrW(sGetTxRxDataIO(cp), (unsigned short *)cbuf, wRecv);
		if (ToRecv & 1)
			cbuf[ToRecv - 1] = sInB(sGetTxRxDataIO(cp));
		memset(fbuf, TTY_NORMAL, ToRecv);
		cbuf += ToRecv;
		fbuf += ToRecv;
		count += ToRecv;
	}

	/*  Push the data up to the tty layer */
#ifdef TTYLAYER_NEW
	tty_flip_buffer_push(tty);
#else
	tty->ldisc.receive_buf(tty, tty->flip.char_buf, tty->flip.flag_buf,count);
#endif

#ifdef TTYLAYER_NEW
done:
	tty_ldisc_deref(ld);
#endif

}//end rp_do_receive()

/*
 *  Serial port transmit data function.  Called from the timer polling loop as a
 *  result of a bit set in xmit_flags[], indicating data (from the tty layer) is ready
 *  to be sent out the serial port.  Data is buffered in rp_table[line].xmit_buf, it is
 *  moved to the port's xmit FIFO.  *info is critical data, protected by spinlocks.
 */
static void rp_do_transmit(struct r_port *info)
{
	int c;
	CHANNEL_t *cp = &info->channel;
	struct tty_struct *tty;

#ifdef ROCKET_DEBUG_INTR
	printk(KERN_INFO "rp_do_transmit ");
#endif

	if (!info)
		return;
	if (!info->tty) {
		printk(KERN_INFO "rp: WARNING rp_do_transmit called with info->tty==NULL\n");
		clear_bit((info->aiop * 8) + info->chan, (void *)&xmit_flags[info->board]);
		return;
	}

	spin_lock_bh(&info->slock);

	tty = info->tty;
	info->xmit_fifo_room = TXFIFO_SIZE - sGetTxCnt(cp);

	/*  Loop sending data to FIFO until done or FIFO full */
	while (1) {
		if (tty->stopped || tty->hw_stopped)
			break;
		c = MIN(info->xmit_fifo_room,
			MIN(info->xmit_cnt, XMIT_BUF_SIZE - info->xmit_tail));
		if (c <= 0 || info->xmit_fifo_room <= 0)
			break;
		sOutStrW(sGetTxRxDataIO(cp), (unsigned short *)(info->xmit_buf + info->xmit_tail), c / 2);
		if (c & 1)
			sOutB(sGetTxRxDataIO(cp), info->xmit_buf[info->xmit_tail + c - 1]);
		info->xmit_tail += c;
		info->xmit_tail &= XMIT_BUF_SIZE - 1;
		info->xmit_cnt -= c;
		info->xmit_fifo_room -= c;
#ifdef ROCKET_DEBUG_INTR
		printk(KERN_INFO "tx %d chars...", c);
#endif
	}//end while

	if (info->xmit_cnt == 0)
		clear_bit((info->aiop * 8) + info->chan, (void *)&xmit_flags[info->board]);

	/*
	 * Calling line discipline wakeup functions while holding a spinlock causes deadlock.
	 */
	spin_unlock_bh(&info->slock);
	if (info->xmit_cnt < WAKEUP_CHARS) {
#ifdef TTYLAYER_NEW
		tty_wakeup(tty);	// +
#else
		if ((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) && tty->ldisc.write_wakeup)
			(tty->ldisc.write_wakeup) (tty);
#endif

		wake_up_interruptible(&tty->write_wait);

#ifdef ROCKETPORT_HAVE_POLL_WAIT /* FIXME: Who ever sets this ? */
		wake_up_interruptible(&tty->poll_wait);
#endif
	}//end if

#ifdef ROCKET_DEBUG_INTR
	printk(KERN_INFO "(%d,%d,%d,%d)...", info->xmit_cnt, info->xmit_head,
	       info->xmit_tail, info->xmit_fifo_room);
#endif

}//end rp_do_transmit()

/*
 *  Called when a serial port signals it has read data in it's RX FIFO.
 *  It checks what interrupts are pending and services them, including
 *  receiving serial data.
 */
static void rp_handle_port(struct r_port *info)
{
	CHANNEL_t *cp;
	struct tty_struct *tty;
	unsigned int IntMask, ChanStatus;

	if (!info)
		return;

	if ((info->flags & ROCKET_INITIALIZED) == 0) {
		printk(KERN_INFO
		       "rp: WARNING: rp_handle_port called with info->flags & NOT_INIT\n");
		return;
	}
	if (!info->tty) {
		//printk(KERN_INFO "rp: WARNING: rp_handle_port %d called with info->tty==NULL\n",info->line);
		return;
	}

	cp = &info->channel;
	tty = info->tty;

	IntMask = sGetChanIntID(cp) & info->intmask;
#ifdef ROCKET_DEBUG_INTR
	printk(KERN_INFO "rp_interrupt %02x...", IntMask);
#endif
	ChanStatus = sGetChanStatus(cp);

	if (IntMask & RXF_TRIG) {	/* Rx FIFO trigger level */
		rp_do_receive(info, tty, cp, ChanStatus);
	}
	if (IntMask & DELTA_CD) {	/* CD change  */
#if (defined(ROCKET_DEBUG_OPEN) || defined(ROCKET_DEBUG_INTR) || defined(ROCKET_DEBUG_HANGUP))
		printk(KERN_INFO "ttyR%d CD now %s...", info->line, (ChanStatus & CD_ACT) ? "on" : "off");
#endif
		if (!(ChanStatus & CD_ACT) && info->cd_status) {
#ifdef ROCKET_DEBUG_HANGUP
			printk(KERN_INFO "CD drop, calling hangup.\n");
#endif
			tty_hangup(tty);
		}
		info->cd_status = (ChanStatus & CD_ACT) ? 1 : 0;
		wake_up_interruptible(&info->open_wait);
	}
#ifdef ROCKET_DEBUG_INTR
	if (IntMask & DELTA_CTS) {	/* CTS change */
		printk(KERN_INFO "CTS change...\n");
	}
	if (IntMask & DELTA_DSR) {	/* DSR change */
		printk(KERN_INFO "DSR change...\n");
	}
#endif
}

/*
 *  The top level polling routine.  Repeats every 1/100 HZ (10ms).
 */
static void rp_do_poll(unsigned long dummy)
{
	CONTROLLER_t *CtlP;
	int ctrl, aiop, ch, line;
	unsigned int xmitmask;
	unsigned int CtlMask;
	unsigned char AiopMask;
	u32 bit;
#if LINUX_VERSION_CODE < VERSION_CODE(2,5,0)
	struct r_port *info;
#endif

	/*  Walk through all the boards (ctrl's) */
	for (ctrl = 0; ctrl < numberof_boards; ctrl++) {

		/*  Get a ptr to the board's control struct */
		CtlP = sCtlNumToCtlPtr(ctrl);

		/*  Get the interupt status from the board */
#ifdef CONFIG_PCI
		if (cardIsPCI(CtlP)) {
			if (CtlP->BaseAiopicIOAddr == 0)
				continue;
			CtlMask = sPCIGetControllerIntStatus(CtlP);
		} else
#endif
		{
			if (CtlP->MBaseIO == 0)
				continue;
			CtlMask = sGetControllerIntStatus(CtlP);
		}

		/*  Check if any AIOP read bits are set */
		for (aiop = 0; CtlMask; aiop++) {
			bit = CtlP->AiopIntrBits[aiop];
			if (CtlMask & bit) {
				CtlMask &= ~bit;
				AiopMask = sGetAiopIntStatus(CtlP, aiop);

				/*  Check if any port read bits are set */
				for (ch = 0; AiopMask; AiopMask >>= 1, ch++) {
					if (AiopMask & 1) {

						/*  Get the line number (/dev/ttyRx number). */
						/*  Read the data from the port. */
						line = GetLineNumber(ctrl, aiop, ch);
						rp_handle_port(rp_table[line]);
					}
				}
			}
		}

		/*
		 *  xmit_flags contains bit-significant flags, indicating there is data
		 *  to xmit on the port. Bit 0 is port 0 on this board, bit 1 is port
		 *  1, ... (32 total possible).
		 */
		xmitmask = xmit_flags[ctrl];

		for (aiop = 0; aiop < 4; aiop++)
			for (ch = 0; ch < 8; ch++) {
				if (xmitmask & 0x00000001)
					rp_do_transmit(rp_table[GetLineNumber(ctrl, aiop, ch)]);
				xmitmask >>= 1;
			}

#if LINUX_VERSION_CODE < VERSION_CODE(2,5,0)
		/*  Walk through the channels, looking for processes sleeping on MSR line changes */
		for (aiop = 0; aiop < 4; aiop++)
			for (ch = 0; ch < 8; ch++) {
				line = GetLineNumber(ctrl, aiop, ch);
				info = rp_table[line];
				if (info && info->am_delta_msr_waiting)
					check_for_delta_msr(info);
			}
#endif
	}

	/*
	 * Reset the timer so we get called at the next clock tick (10ms).
	 */
	if (atomic_read(&rp_num_ports_open))
		mod_timer(&rocket_timer, jiffies + POLL_PERIOD);
}

/*
 *  Initializes the r_port structure for a port, as well as enabling the port on
 *  the board.
 *  Inputs:  board, aiop, chan numbers
 */
static void init_r_port(int board, int aiop, int chan, struct pci_dev *pci_dev)
{
	unsigned rocketMode;
	struct r_port *info;
	int line;
	CONTROLLER_T *CtlP;

	/*  Get the next available line number */
	line = SetLineNumber(board, aiop, chan);

	CtlP = sCtlNumToCtlPtr(board);
#ifdef ROCKET_DEBUG_IO
	if (!CtlP) {
		printk("NULL CtlP in init_r_port\n");
		return;
	}
#endif

	/*  Get a r_port struct for the port, fill it in and save it globally, indexed by line number */
	info = kmalloc(sizeof(struct r_port), GFP_KERNEL);
	if (!info) {
		printk(KERN_INFO "Couldn't allocate info struct for line #%d\n",
		       line);
		return;
	}
	memset(info, 0, sizeof(struct r_port));

	info->magic = RPORT_MAGIC;
	info->line = line;
	info->CtlP = CtlP;
	info->board = board;
	info->aiop = aiop;
	info->chan = chan;
	info->closing_wait = 3000;
	info->close_delay = 50;
#if LINUX_VERSION_CODE < VERSION_CODE(2,5,0)
	info->normal_termios = rocket_driver->init_termios;
#endif

	/* initialize our wait queues */
	init_waitqueue_head(&info->open_wait);
	init_waitqueue_head(&info->close_wait);
#if LINUX_VERSION_CODE < VERSION_CODE(2,5,0)
	init_waitqueue_head(&info->delta_msr_wait);
#endif

	/* initialize our icount structure */
#if LINUX_VERSION_CODE < VERSION_CODE(2,5,0)
	memset(&(info->icount), 0x00, sizeof(info->icount));
	info->am_delta_msr_waiting = 0;
#endif

	info->flags &= ~ROCKET_MODE_MASK;
	switch (pc104[board][line]) {
	case 422:
		info->flags |= ROCKET_MODE_RS422;
		break;
	case 485:
		info->flags |= ROCKET_MODE_RS485;
		break;
	case 232:
	default:
		info->flags |= ROCKET_MODE_RS232;
		break;
	}

	info->intmask =
		RXF_TRIG | TXFIFO_MT | SRC_INT | DELTA_CD | DELTA_CTS | DELTA_DSR;
	if (sInitChan(CtlP, &info->channel, aiop, chan) == 0) {
		printk(KERN_INFO "RocketPort sInitChan(%d, %d, %d) failed!\n",
		       board, aiop, chan);
		kfree(info);
		return;
	}

	rocketMode = info->flags & ROCKET_MODE_MASK;

	if ((info->flags & ROCKET_RTS_TOGGLE)
	    || (rocketMode == ROCKET_MODE_RS485))
		sEnRTSToggle(&info->channel);
	else
		sDisRTSToggle(&info->channel);

	if (isa_type[board] == ROCKET_TYPE_PC104) {
		switch (rocketMode) {
		case ROCKET_MODE_RS485:
			sSetInterfaceMode(&info->channel, InterfaceModeRS485);
			break;
		case ROCKET_MODE_RS422:
			sSetInterfaceMode(&info->channel, InterfaceModeRS422);
			break;
		case ROCKET_MODE_RS232:
		default:
			if (info->flags & ROCKET_RTS_TOGGLE)
				sSetInterfaceMode(&info->channel,
						  InterfaceModeRS232T);
			else
				sSetInterfaceMode(&info->channel,
						  InterfaceModeRS232);
			break;
		}
	}
	spin_lock_init(&info->slock);
#if LINUX_VERSION_CODE < VERSION_CODE(2,6,10)
	init_MUTEX(&info->tmp_buf_sem);
#endif

	rp_table[line] = info;
#if LINUX_VERSION_CODE > VERSION_CODE(2,5,0)
	if (pci_dev)
		tty_register_device(rocket_driver, line, &pci_dev->dev);
#endif

#ifdef ROCKET_DEBUG_IO
	printk(KERN_INFO
	       "Configured channel -  AiopNum: %d  ChanID: %d  ChanNum: %d Cmd: 0x%x IntChan: 0x%x IntMask: 0x%x \n",
	       info->channel.AiopNum, info->channel.ChanID,
	       info->channel.ChanNum, info->channel.Cmd, info->channel.IntChan,
	       info->channel.IntMask);
#endif
}

/*
 *  Configures a rocketport port according to its termio settings.  Called from
 *  user mode into the driver (exception handler).  *info CD manipulation is spinlock protected.
 */
#if LINUX_VERSION_CODE < VERSION_CODE(2,6,20)
static void configure_r_port(struct r_port *info, struct termios *old_termios)
#else
static void configure_r_port(struct r_port *info, struct ktermios *old_termios)
#endif
{
	unsigned cflag;
	unsigned rocketMode;
	int bits, baud, divisor;
	CHANNEL_t *cp;

	if (!info->tty || !info->tty->termios)
		return;
	cp = &info->channel;
	cflag = info->tty->termios->c_cflag;

	/* Byte size and parity */
	if ((cflag & CSIZE) == CS8) {
		sSetData8(cp);
		bits = 10;
	} else {
		sSetData7(cp);
		bits = 9;
	}

	if (cflag & CSTOPB) {
		sSetStop2(cp);
		bits++;
	} else
		sSetStop1(cp);

	if (cflag & PARENB) {
		sEnParity(cp);
		bits++;
		if (cflag & PARODD)
			sSetOddParity(cp);
		else
			sSetEvenParity(cp);
	} else
		sDisParity(cp);

	/* baud rate */
	baud = tty_get_baud_rate(info->tty);
	if (!baud)
		baud = 9600;
	divisor = ((info->CtlP->rp_baud_base + (baud >> 1)) / baud) - 1;
	if ((divisor >= 8192 || divisor < 0) && old_termios) {
		info->tty->termios->c_cflag &= ~CBAUD;
		info->tty->termios->c_cflag |= (old_termios->c_cflag & CBAUD);
		baud = tty_get_baud_rate(info->tty);
		if (!baud)
			baud = 9600;
		divisor = (info->CtlP->rp_baud_base / baud) - 1;
	}

	if (divisor >= 8192 || divisor < 0) {
		baud = 9600;
		divisor = (info->CtlP->rp_baud_base / baud) - 1;
	}
	info->cps = baud / bits;
	sSetBaud(cp, divisor);

	if (cflag & CRTSCTS) {
		info->intmask |= DELTA_CTS;
		sEnCTSFlowCtl(cp);
	} else {
		info->intmask &= ~DELTA_CTS;
		sDisCTSFlowCtl(cp);
	}

	if (cflag & CLOCAL) {
		info->intmask &= ~DELTA_CD;
	} else {
		spin_lock_bh(&info->slock);
		if (sGetChanStatus(cp) & CD_ACT)
			info->cd_status = 1;
		else
			info->cd_status = 0;
		info->intmask |= DELTA_CD;
		spin_unlock_bh(&info->slock);
	}

	/*
	 * Handle software flow control in the board
	 */
#ifdef ROCKET_SOFT_FLOW
	if (I_IXON(info->tty)) {
		sEnTxSoftFlowCtl(cp);
		if (I_IXANY(info->tty))
			sEnIXANY(cp);
		else
			sDisIXANY(cp);
		sSetTxXONChar(cp, START_CHAR(info->tty));
		sSetTxXOFFChar(cp, STOP_CHAR(info->tty));
	} else {
		sDisTxSoftFlowCtl(cp);
		sDisIXANY(cp);
		sClrTxXOFF(cp);
	}
#endif

	/*
	 * Set up ignore/read mask words
	 */
	info->read_status_mask = STMRCVROVRH | 0xFF;
	if (I_INPCK(info->tty))
		info->read_status_mask |= STMFRAMEH | STMPARITYH;
	if (I_BRKINT(info->tty) || I_PARMRK(info->tty))
		info->read_status_mask |= STMBREAKH;

	/*
	 * Characters to ignore
	 */
	info->ignore_status_mask = 0;
	if (I_IGNPAR(info->tty))
		info->ignore_status_mask |= STMFRAMEH | STMPARITYH;
	if (I_IGNBRK(info->tty)) {
		info->ignore_status_mask |= STMBREAKH;
		/*
		 * If we're ignoring parity and break indicators,
		 * ignore overruns too.  (For real raw support).
		 */
		if (I_IGNPAR(info->tty))
			info->ignore_status_mask |= STMRCVROVRH;
	}

	rocketMode = info->flags & ROCKET_MODE_MASK;

	if ((info->flags & ROCKET_RTS_TOGGLE)
	    || (rocketMode == ROCKET_MODE_RS485))
		sEnRTSToggle(cp);
	else
		sDisRTSToggle(cp);

	sSetRTS(&info->channel);

	if (cardIsPC104(info->CtlP)) {
		switch (rocketMode) {
		case ROCKET_MODE_RS485:
			sSetInterfaceMode(cp, InterfaceModeRS485);
			break;
		case ROCKET_MODE_RS422:
			sSetInterfaceMode(cp, InterfaceModeRS422);
			break;
		case ROCKET_MODE_RS232:
		default:
			if (info->flags & ROCKET_RTS_TOGGLE)
				sSetInterfaceMode(cp, InterfaceModeRS232T);
			else
				sSetInterfaceMode(cp, InterfaceModeRS232);
			break;
		}
	}
}

/*  info->count is considered critical, protected by spinlocks.  */
static int block_til_ready(struct tty_struct *tty, struct file *filp, struct r_port *info)
{
	DECLARE_WAITQUEUE(wait, current);
	int retval;
	int do_clocal = 0, extra_count = 0;

	/*
	 * If the device is in the middle of being closed, then block
	 * until it's done, and then try again.
	 */
	if (tty_hung_up_p(filp))
		return ((info->flags & ROCKET_HUP_NOTIFY) ? -EAGAIN : -ERESTARTSYS);

	if (info->flags & ROCKET_CLOSING) {
		wait_event_interruptible(info->close_wait,(!(info->flags&ROCKET_CLOSING)));
		return ((info->flags & ROCKET_HUP_NOTIFY) ? -EAGAIN : -ERESTARTSYS);
	}
	/*
	 * If non-blocking mode is set, or the port is not enabled,
	 * then make the check up front and then exit.
	 */
	if ((filp->f_flags & O_NONBLOCK) || (tty->flags & (1 << TTY_IO_ERROR))) {
		info->flags |= ROCKET_NORMAL_ACTIVE;
		return 0;
	}
	if (tty->termios->c_cflag & CLOCAL)
		do_clocal = 1;

	/*
	 * Block waiting for the carrier detect and the line to become free.  While we are in
	 * this loop, info->count is dropped by one, so that rp_close() knows when to free things.
	 * We restore it upon exit, either normal or abnormal.
	 */
	retval = 0;
	add_wait_queue(&info->open_wait, &wait);
#ifdef ROCKET_DEBUG_OPEN
	printk(KERN_INFO "block_til_ready before block: ttyR%d, count = %d\n", info->line, info->count);
#endif
	spin_lock_bh(&info->slock);

#ifdef ROCKET_DISABLE_SIMUSAGE
	info->flags |= ROCKET_NORMAL_ACTIVE;
#else
	if (!tty_hung_up_p(filp)) {
		extra_count = 1;
		info->count--;
	}
#endif
	info->blocked_open++;

	spin_unlock_bh(&info->slock);

	while (1) {
		if (tty->termios->c_cflag & CBAUD) {
			sSetDTR(&info->channel);
			sSetRTS(&info->channel);
		}
		set_current_state(TASK_INTERRUPTIBLE);
		if (tty_hung_up_p(filp) || !(info->flags & ROCKET_INITIALIZED)) {
			if (info->flags & ROCKET_HUP_NOTIFY)
				retval = -EAGAIN;
			else
				retval = -ERESTARTSYS;
			break;
		}
		if (!(info->flags & ROCKET_CLOSING) &&
		    (do_clocal	|| (sGetChanStatusLo(&info->channel) & CD_ACT)))
			break;
		if (signal_pending(current)) {
			retval = -ERESTARTSYS;
			break;
		}
#ifdef ROCKET_DEBUG_OPEN
		printk(KERN_INFO "block_til_ready blocking: ttyR%d, count = %d, flags=0x%0x\n", info->line, info->count, info->flags);
#endif
		schedule();	/*  Don't hold spinlock here, will hang PC */
	}
	current->state = TASK_RUNNING;
	remove_wait_queue(&info->open_wait, &wait);

	spin_lock_bh(&info->slock);

	if (extra_count)
		info->count++;

	info->blocked_open--;

	spin_unlock_bh(&info->slock);

#ifdef ROCKET_DEBUG_OPEN
	printk(KERN_INFO "block_til_ready after blocking: ttyR%d, count = %d\n", info->line, info->count);
#endif
	if (retval)
		return retval;

	info->flags |= ROCKET_NORMAL_ACTIVE;

	return 0;
}//end block_til_ready()

/*
 *  Exception handler that opens a serial port.  Creates xmit_buf storage, fills in
 *  port's r_port struct.  Initializes the port hardware.
 */
static int rp_open(struct tty_struct *tty, struct file *filp)
{
	struct r_port *info;
	int line = 0, retval;
	CHANNEL_t *cp;
	unsigned long pagex;

//compiler complains if this isn't conditional
#if LINUX_VERSION_CODE < VERSION_CODE(2,6,10)
	unsigned long paget;
#endif

	line = TTY_GET_LINE(tty);

	if ((line < 0) || (line >= MAX_PORTS) || ((info = rp_table[line]) == NULL))
		return -ENXIO;

	pagex = __get_free_page(GFP_KERNEL);
	if (!pagex)
		return -ENOMEM;

#if LINUX_VERSION_CODE < VERSION_CODE(2,6,10)
	paget = __get_free_page(GFP_KERNEL);
	if (!paget) {
		free_page(pagex);
		return -ENOMEM;
	}
#endif

	if (info->flags & ROCKET_CLOSING) {
		wait_event_interruptible(info->close_wait,(!(info->flags&ROCKET_CLOSING)));
		free_page(pagex);
#if LINUX_VERSION_CODE < VERSION_CODE(2,6,10)
		free_page(paget);
#endif
		return ((info->flags & ROCKET_HUP_NOTIFY) ? -EAGAIN : -ERESTARTSYS);
	}

	/*
	 * We must not sleep from here until the port is marked fully in use.
	 */
	if (info->xmit_buf)
		free_page(pagex);
	else
		info->xmit_buf = (unsigned char *)pagex;

#if LINUX_VERSION_CODE < VERSION_CODE(2,6,10)
	if (info->tmp_buf)
		free_page(paget);
	else
		info->tmp_buf = (unsigned char *)paget;
#endif

	tty->driver_data = info;
	info->tty = tty;

	if (info->count++ == 0) {
#if LINUX_VERSION_CODE < VERSION_CODE(2,5,0)
#if defined(MODULE)
		MOD_INC_USE_COUNT;
#endif
#endif
		atomic_inc(&rp_num_ports_open);

#ifdef ROCKET_DEBUG_OPEN
		printk(KERN_INFO "rocket mod++ = %d \n",
		       atomic_read(&rp_num_ports_open));
#endif
	}

#ifdef ROCKET_DEBUG_OPEN
	printk(KERN_INFO "rp_open ttyR%d, count=%d \n", info->line, info->count);
#endif

	/*
	 * Info->count is now 1; so it's safe to sleep now.
	 */
	if ((info->flags & ROCKET_INITIALIZED) == 0) {
		cp = &info->channel;
		sSetRxTrigger(cp, TRIG_1);
		if (sGetChanStatus(cp) & CD_ACT)
			info->cd_status = 1;
		else
			info->cd_status = 0;
		sDisRxStatusMode(cp);
		sFlushRxFIFO(cp);
		sFlushTxFIFO(cp);

		sSetRxTrigger(cp, TRIG_1);

		sGetChanStatus(cp);
		sDisRxStatusMode(cp);
		sClrTxXOFF(cp);

		sDisCTSFlowCtl(cp);
		sDisTxSoftFlowCtl(cp);

		sEnRxFIFO(cp);
		sEnTransmit(cp);

		info->flags |= ROCKET_INITIALIZED;

		sEnInterrupts(cp, (MCINT_EN | RXINT_EN | SRCINT_EN | CHANINT_EN));

		/*
		 * Set up the tty->alt_speed kludge
		 */
		if ((info->flags & ROCKET_SPD_MASK) == ROCKET_SPD_HI)
			info->tty->alt_speed = 57600;
		if ((info->flags & ROCKET_SPD_MASK) == ROCKET_SPD_VHI)
			info->tty->alt_speed = 115200;
		if ((info->flags & ROCKET_SPD_MASK) == ROCKET_SPD_SHI)
			info->tty->alt_speed = 230400;
		if ((info->flags & ROCKET_SPD_MASK) == ROCKET_SPD_WARP)
			info->tty->alt_speed = 460800;
		if ((info->flags & ROCKET_SPD_MASK) == ROCKET_SPD_WARP13)
			info->tty->alt_speed = 921600;

		configure_r_port(info, NULL);
		if (tty->termios->c_cflag & CBAUD) {
			sSetDTR(cp);
			sSetRTS(cp);
		}
	}
	/*  Starts (or resets) the maint polling loop */
	mod_timer(&rocket_timer, jiffies + POLL_PERIOD);

	retval = block_til_ready(tty, filp, info);
	if (retval) {
#ifdef ROCKET_DEBUG_OPEN
		printk(KERN_INFO "rp_open returning after block_til_ready with %d\n", retval);
#endif
		return retval;
	}
#if LINUX_VERSION_CODE < VERSION_CODE(2,5,0)
	if ((info->count == 1) && (info->flags & ROCKET_SPLIT_TERMIOS)) {
		*tty->termios = info->normal_termios;
		configure_r_port(info, NULL);
	}
#endif
	return 0;
}

/*
 *  Exception handler that closes a serial port. info->count is considered critical.
 */
static void rp_close(struct tty_struct *tty, struct file *filp)
{
	struct r_port *info = (struct r_port *)tty->driver_data;
	int timeout;
	CHANNEL_t *cp;

	if (rocket_paranoia_check(info, "rp_close"))
		return;

#ifdef ROCKET_DEBUG_OPEN
	printk(KERN_INFO "rp_close ttyR%d, count = %d\n", info->line, info->count);
#endif

	if (tty_hung_up_p(filp))
		return;

	spin_lock_bh(&info->slock);

	if (--info->count < 0) {
		printk(KERN_INFO "rp_close: bad serial port count for ttyR%d: %d\n", info->line, info->count);
		info->count = 0;
	}
	if (info->count) {
		spin_unlock_bh(&info->slock);
		return;
	}
	info->flags |= ROCKET_CLOSING;
	spin_unlock_bh(&info->slock);

#if LINUX_VERSION_CODE < VERSION_CODE(2,5,0)
	info->normal_termios = *tty->termios;
#endif
	cp = &info->channel;

	/*
	 *  If a process is sleeping on TIOCMIWAIT, wake it up now.
	 */
#if LINUX_VERSION_CODE < VERSION_CODE(2,5,0)
	wake_up_interruptible(&info->delta_msr_wait);
#endif

	/*
	 * Notify the line discpline to only process XON/XOFF characters
	 */
	tty->closing = 1;

	/*
	 * If transmission was throttled by the application request,
	 * just flush the xmit buffer.
	 */
	if (tty->flow_stopped)
		rp_flush_buffer(tty);

	/*
	 * Wait for the transmit buffer to clear
	 */
	if (info->closing_wait != ROCKET_CLOSING_WAIT_NONE)
		tty_wait_until_sent(tty, info->closing_wait);
	/*
	 * Before we drop DTR, make sure the UART transmitter
	 * has completely drained; this is especially
	 * important if there is a transmit FIFO!
	 */
	timeout = (sGetTxCnt(cp) + 1) * HZ / info->cps;
	if (timeout == 0)
		timeout = 1;
	rp_wait_until_sent(tty, timeout);
	clear_bit((info->aiop * 8) + info->chan,
		  (void *)&xmit_flags[info->board]);

	sDisTransmit(cp);
	sDisInterrupts(cp,
		       (TXINT_EN | MCINT_EN | RXINT_EN | SRCINT_EN |
			CHANINT_EN));
	sDisCTSFlowCtl(cp);
	sDisTxSoftFlowCtl(cp);
	sClrTxXOFF(cp);
	sFlushRxFIFO(cp);
	sFlushTxFIFO(cp);
	sClrRTS(cp);
	if (C_HUPCL(tty))
		sClrDTR(cp);

	if (TTY_DRIVER_FLUSH_BUFFER_EXISTS(tty))
		TTY_DRIVER_FLUSH_BUFFER(tty);
#if LINUX_VERSION_CODE < VERSION_CODE(2,6,27)
	if (tty->ldisc.flush_buffer)
		tty->ldisc.flush_buffer(tty);
#elif LINUX_VERSION_CODE < VERSION_CODE(2,6,31)
	if (tty->ldisc.ops->flush_buffer)
		tty->ldisc.ops->flush_buffer(tty);
#else	// implies kernel >= 2,6,31
	if (tty->ldisc->ops->flush_buffer)
		tty->ldisc->ops->flush_buffer(tty);
#endif

	clear_bit((info->aiop * 8) + info->chan,
		  (void *)&xmit_flags[info->board]);

	if (info->blocked_open) {
		if (info->close_delay) {
			current->state = TASK_INTERRUPTIBLE;
			schedule_timeout(info->close_delay);
		}
		wake_up_interruptible(&info->open_wait);
	} else {
		if (info->xmit_buf) {
			free_page((unsigned long)info->xmit_buf);
			info->xmit_buf = 0;
		}
#if LINUX_VERSION_CODE < VERSION_CODE(2,6,10)
		if (info->tmp_buf) {
			free_page((unsigned long)info->tmp_buf);
			info->tmp_buf = 0;
		}
#endif
	}
	info->flags &= ~(ROCKET_INITIALIZED | ROCKET_CLOSING | ROCKET_NORMAL_ACTIVE);
	tty->closing = 0;
	wake_up_interruptible(&info->close_wait);

#if LINUX_VERSION_CODE < VERSION_CODE(2,5,0)
#if defined(MODULE)
	MOD_DEC_USE_COUNT;
#endif
#endif

	atomic_dec(&rp_num_ports_open);

#ifdef ROCKET_DEBUG_OPEN
	printk(KERN_INFO "rocket mod-- = %d...",
	       atomic_read(&rp_num_ports_open));
	printk(KERN_INFO "rp_close ttyR%d complete shutdown\n", info->line);
#endif

}//end rp_close

#if LINUX_VERSION_CODE < VERSION_CODE(2,6,20)
static void rp_set_termios(struct tty_struct *tty, struct termios *old_termios)
#else
static void rp_set_termios(struct tty_struct *tty, struct ktermios *old_termios)
#endif
{
	struct r_port *info = (struct r_port *)tty->driver_data;
	CHANNEL_t *cp;
	unsigned cflag;

	if (rocket_paranoia_check(info, "rp_set_termios"))
		return;

	cflag = tty->termios->c_cflag;

	if (cflag == old_termios->c_cflag)
		return;

	/*
	 * This driver doesn't support CS5 or CS6
	 */
	if (((cflag & CSIZE) == CS5) || ((cflag & CSIZE) == CS6))
		tty->termios->c_cflag =
			((cflag & ~CSIZE) | (old_termios->c_cflag & CSIZE));

	configure_r_port(info, old_termios);

	cp = &info->channel;

	/* Handle transition to B0 status */
	if ((old_termios->c_cflag & CBAUD) && !(tty->termios->c_cflag & CBAUD)) {
		sClrDTR(cp);
		sClrRTS(cp);
	}

	/* Handle transition away from B0 status */
	if (!(old_termios->c_cflag & CBAUD) && (tty->termios->c_cflag & CBAUD)) {
		if (!tty->hw_stopped || !(tty->termios->c_cflag & CRTSCTS))
			sSetRTS(cp);
		sSetDTR(cp);
	}

	if ((old_termios->c_cflag & CRTSCTS)
	    && !(tty->termios->c_cflag & CRTSCTS)) {
		tty->hw_stopped = 0;
		rp_start(tty);
	}
}

#if LINUX_VERSION_CODE < VERSION_CODE(2,6,27)
static void rp_break(struct tty_struct *tty, int break_state)
#else
static int rp_break(struct tty_struct *tty, int break_state)
#endif
{
	struct r_port *info = (struct r_port *)tty->driver_data;

	if (rocket_paranoia_check(info, "rp_break"))
#if LINUX_VERSION_CODE < VERSION_CODE(2,6,27)
		return;
#else
		return 0;
#endif

	spin_lock_bh(&info->slock);
	if (break_state == -1)
		sSendBreak(&info->channel);
	else
		sClrBreak(&info->channel);
	spin_unlock_bh(&info->slock);
#if LINUX_VERSION_CODE >= VERSION_CODE(2,6,27)
	return 0;
#endif
}

/*
 *  Returns the state of RI. The following products have RI:  UPCI 4/8 port,
 *  RMIII 4/8 port, RMIV 4/8 port, RP Plus 4/8 port, RP Plus 2 port 232/422, PC 104.
 *  For all other devices, this function returns 0.
 *
 *  Returns:  1 if RI is set, else 0.
 */
static int sGetChanRI(CHANNEL_T * ChP)
{
	CONTROLLER_t *CtlP = ChP->CtlP;
	int ChanNum = ChP->ChanNum;
	int RingInd = 0;

	if (cardIsUPCI(CtlP) && !cardIsRocketportPlus(CtlP))
		RingInd =
			!(sInB
			  (CtlP->BaseLocalConfigurationIOAddr +
			   _PCI_9030_RING_IND) & sBitMapSetTbl[ChanNum]);
	else if (cardIsRocketportPlus(CtlP) || cardIsURPPluswRI(CtlP))
		RingInd = sInB((u32) (ChP->ChanStat + 8)) & DSR_ACT;
	else if (cardIsPC104(CtlP))
		RingInd = !(sInB(CtlP->AiopIO[3]) & sBitMapSetTbl[ChanNum]);

	return RingInd;
}

/********************************************************************************************/
/*  Here are the routines used by rp_ioctl.  These are all called from exception handlers.  */

#if LINUX_VERSION_CODE < VERSION_CODE(2,5,0)
/*
 *  Gets the MSR values for the port
 */
static unsigned long get_msr(struct r_port *info)
{
#else
/*
 *  Returns the state of the serial modem control lines.  These next 2 functions
 *  are the way kernel versions > 2.5 handle modem control lines rather than IOCTLs.
 */
#if LINUX_VERSION_CODE < VERSION_CODE(2,6,39)
static int rp_tiocmget(struct tty_struct *tty, struct file *file)
#else
	static int rp_tiocmget(struct tty_struct *tty)
#endif
{
	struct r_port *info = (struct r_port *)tty->driver_data;
#endif
	unsigned int control, result, ChanStatus;

	ChanStatus = sGetChanStatusLo(&info->channel);
	control = info->channel.TxControl[3];
	result = ((control & SET_RTS) ? TIOCM_RTS : 0) |
		 ((control & SET_DTR) ? TIOCM_DTR : 0) |
		 ((ChanStatus & CD_ACT) ? TIOCM_CAR : 0) |
		 (sGetChanRI(&info->channel) ? TIOCM_RNG : 0) |
		 ((ChanStatus & DSR_ACT) ? TIOCM_DSR : 0) |
		 ((ChanStatus & CTS_ACT) ? TIOCM_CTS : 0);
	return result;
}

#if LINUX_VERSION_CODE > VERSION_CODE(2,5,0)
/*
 *  Sets the modem control lines
 */
#if LINUX_VERSION_CODE < VERSION_CODE(2,6,39)
static int rp_tiocmset(struct tty_struct *tty, struct file *file, unsigned int set, unsigned int clear)
#else
static int rp_tiocmset(struct tty_struct *tty, unsigned int set, unsigned int clear)
#endif
{
	struct r_port *info = (struct r_port *)tty->driver_data;

	if (set & TIOCM_RTS)
		info->channel.TxControl[3] |= SET_RTS;
	if (set & TIOCM_DTR)
		info->channel.TxControl[3] |= SET_DTR;
	if (clear & TIOCM_RTS)
		info->channel.TxControl[3] &= ~SET_RTS;
	if (clear & TIOCM_DTR)
		info->channel.TxControl[3] &= ~SET_DTR;
	sOutDW(info->channel.IndexAddr, *(u32 *) & (info->channel.TxControl[0]));
	return 0;
}

#endif				/*  Linux > 2.5 */

/*
 *  Gets the MSR values for the port, returns it to userspace through the userAddr ptr.
 */
static int get_modem_info(struct r_port *info, unsigned long userAddr)
{
	unsigned int result;

#if LINUX_VERSION_CODE < VERSION_CODE(2,5,0)
	result = get_msr(info);
#endif
	if (copy_to_user((void *)userAddr, &result, sizeof(int)))
		return -EFAULT;
	return 0;
}

static int set_modem_info(struct r_port *info, unsigned int cmd, unsigned long userAddr)
{
	unsigned int arg;

	if (copy_from_user(&arg, (void *)userAddr, sizeof(int)))
		return -EFAULT;

	switch (cmd) {
	case TIOCMBIS:
		if (arg & TIOCM_RTS)
			info->channel.TxControl[3] |= SET_RTS;
		if (arg & TIOCM_DTR)
			info->channel.TxControl[3] |= SET_DTR;
		break;
	case TIOCMBIC:
		if (arg & TIOCM_RTS)
			info->channel.TxControl[3] &= ~SET_RTS;
		if (arg & TIOCM_DTR)
			info->channel.TxControl[3] &= ~SET_DTR;
		break;
	case TIOCMSET:
		info->channel.TxControl[3] =
			((info->channel.
			  TxControl[3] & ~(SET_RTS | SET_DTR)) | ((arg & TIOCM_RTS)
					  ? SET_RTS : 0) |
			 ((arg & TIOCM_DTR) ? SET_DTR : 0));
		break;
	default:
		return -EINVAL;
	}
	sOutDW(info->channel.IndexAddr, *(u32 *) & (info->channel.TxControl[0]));
	return 0;
}

static int get_config(struct r_port *info, struct rocket_config *retinfo)
{
	struct rocket_config tmp;

	if (!retinfo)
		return -EFAULT;
	memset(&tmp, 0, sizeof(tmp));
	tmp.line = info->line;
	tmp.flags = info->flags;
	tmp.close_delay = info->close_delay;
	tmp.closing_wait = info->closing_wait;
	tmp.port = info->CtlP->BaseAiopicIOAddr;

	if (copy_to_user(retinfo, &tmp, sizeof(*retinfo)))
		return -EFAULT;
	return 0;
}

static int set_config(struct r_port *info, struct rocket_config *new_info)
{
	struct rocket_config new_serial;

	if (copy_from_user(&new_serial, new_info, sizeof(new_serial)))
		return -EFAULT;

#ifdef CAP_SYS_ADMIN
	if (!capable(CAP_SYS_ADMIN))
#else
	if (!suser())
#endif
	{
		if ((new_serial.flags & ~ROCKET_USR_MASK) !=
		    (info->flags & ~ROCKET_USR_MASK))
			return -EPERM;
		info->flags =
			((info->flags & ~ROCKET_USR_MASK) | (new_serial.
					flags &
					ROCKET_USR_MASK));
		configure_r_port(info, 0);
		return 0;
	}

	info->flags = ((info->flags & ~ROCKET_FLAGS) | (new_serial.flags & ROCKET_FLAGS));
	info->close_delay = new_serial.close_delay;
	info->closing_wait = new_serial.closing_wait;

	if ((info->flags & ROCKET_SPD_MASK) == ROCKET_SPD_HI)
		info->tty->alt_speed = 57600;
	if ((info->flags & ROCKET_SPD_MASK) == ROCKET_SPD_VHI)
		info->tty->alt_speed = 115200;
	if ((info->flags & ROCKET_SPD_MASK) == ROCKET_SPD_SHI)
		info->tty->alt_speed = 230400;
	if ((info->flags & ROCKET_SPD_MASK) == ROCKET_SPD_WARP)
		info->tty->alt_speed = 460800;
	if ((info->flags & ROCKET_SPD_MASK) == ROCKET_SPD_WARP13)
		info->tty->alt_speed = 921600;

	configure_r_port(info, 0);
	return 0;
}

/*
 *  This function fills in a rocket_ports struct with information
 *  about what boards/ports are in the system.  This info is passed
 *  to user space.  See setrocket.c where the info is used to create
 *  the /dev/ttyRx ports.
 */
static int get_ports(struct r_port *info, struct rocket_ports *retports)
{
	struct rocket_ports tmp;
	int board;

	if (!retports)
		return -EFAULT;
	memset(&tmp, 0, sizeof(tmp));
	tmp.tty_major = rocket_driver->major;

	for (board = 0; board < 4; board++) {
		tmp.rocketModel[board].model =
			(sCtlNumToCtlPtr(board))->rocketModel.model;
		strcpy(tmp.rocketModel[board].modelString,
		       (sCtlNumToCtlPtr(board))->rocketModel.modelString);
		tmp.rocketModel[board].numPorts =
			(sCtlNumToCtlPtr(board))->rocketModel.numPorts;
		tmp.rocketModel[board].loadrm2 =
			(sCtlNumToCtlPtr(board))->rocketModel.loadrm2;
		tmp.rocketModel[board].startingPortNumber =
			(sCtlNumToCtlPtr(board))->rocketModel.startingPortNumber;
	}
	if (copy_to_user(retports, &tmp, sizeof(*retports)))
		return -EFAULT;
	return 0;
}

static int reset_rm2(struct r_port *info, unsigned long arg)
{
	int reset;

	if (copy_from_user(&reset, (void *)arg, sizeof(int)))
		return -EFAULT;

	if (reset)
		reset = 1;

	if (!cardIsRM2(info->CtlP) && !cardIsRM3(info->CtlP) && !cardIsRM4(info->CtlP))
		return -EINVAL;

	sPCIModemReset(info->CtlP, info->chan, reset);
	return 0;
}

static int get_version(struct r_port *info, struct rocket_version *retvers)
{
	if (copy_to_user(retvers, &driver_version, sizeof(*retvers)))
		return -EFAULT;
	return 0;
}

/*
 *  Wait for any of the 4 modem inputs (DCD,RI,DSR,CTS) to change.
 *  Mask passed in arg for lines of interest (use |'ed TIOCM_RNG/DSR/CD/CTS
 *  for masking).  Caller should use TIOCGICOUNT to see which one it was.
 *  Inputs:  ptr to the ports info struct, arg (mask of MSR lines of interest).
 *  Returns: 0 if OK, errno else
 */
#if LINUX_VERSION_CODE < VERSION_CODE(2,5,0)
static int wait_msr_change(struct r_port *info, unsigned long *arg)
{
	DECLARE_WAITQUEUE(wait, current);
	struct async_icount cprev, cnow;
	int ret;
	unsigned char mask = (unsigned char)*arg;

#ifdef ROCKET_DEBUG_DELTA_MSR
	printk(KERN_INFO "Entering msr_wait() - arg: %p  mask: 0x%x \n", arg,
	       mask);
#endif

	/* save current irq counts */
	spin_lock_bh(&info->slock);
	cprev = info->icount;
	spin_unlock_bh(&info->slock);
	add_wait_queue(&info->delta_msr_wait, &wait);

	for (;;) {
		info->am_delta_msr_waiting = 1;
		set_current_state(TASK_INTERRUPTIBLE);
		schedule();
		info->am_delta_msr_waiting = 0;
		if (signal_pending(current)) {
			ret = -ERESTARTSYS;
			break;
		}

		/* get new irq counts */
		spin_lock_bh(&info->slock);
		cnow = info->icount;
		spin_unlock_bh(&info->slock);

		/* if no change, wait aborted for some reason */
		if (cnow.rng == cprev.rng && cnow.dsr == cprev.dsr &&
		    cnow.dcd == cprev.dcd && cnow.cts == cprev.cts) {
			ret = -EIO;
#ifdef ROCKET_DEBUG_DELTA_MSR
			printk(KERN_INFO
			       "msr_wait() woken however no change detected \n");
#endif
			break;
		}

		/* check for change in caller specified modem input */
		if ((mask & TIOCM_RNG && cnow.rng != cprev.rng) ||
		    (mask & TIOCM_DSR && cnow.dsr != cprev.dsr) ||
		    (mask & TIOCM_CD && cnow.dcd != cprev.dcd) ||
		    (mask & TIOCM_CTS && cnow.cts != cprev.cts)) {
			ret = 0;
#ifdef ROCKET_DEBUG_DELTA_MSR
			printk(KERN_INFO
			       "msr_wait() woken - change detected \n");
#endif
			break;
		}

		/*
		 *  This condition happens if we had a change, but it
		 *  wasnt' the one the user asked for in the mask. Loop.
		 */
		cprev = cnow;
#ifdef ROCKET_DEBUG_DELTA_MSR
		printk(KERN_INFO
		       "msr_wait() woken - change detected not in user mask, going back to sleep...\n");
#endif
	}

	remove_wait_queue(&info->delta_msr_wait, &wait);
	set_current_state(TASK_RUNNING);

#ifdef ROCKET_DEBUG_DELTA_MSR
	printk(KERN_INFO "Exiting msr_wait() \n");
#endif

	return ret;
}
#endif				/* wait_msr_change 2.5.0 check */
/*
 *  Get counter of input serial line "interrupts" (changes - DCD,RI,DSR,CTS), returns
 *  data to the userspace.  Counting only occurs if a process is sleeping, waiting to
 *  be woken up by using the TIOCMIWAIT ioctl.
 */
#if LINUX_VERSION_CODE < VERSION_CODE(2,5,0)
static unsigned long get_delta_msr_counts(struct r_port *info,
		struct serial_icounter_struct
		*user_icount)
{
	struct serial_icounter_struct icount;
	struct async_icount cnow;

	spin_lock_bh(&info->slock);
	cnow = info->icount;
	spin_unlock_bh(&info->slock);

	icount.cts = cnow.cts;
	icount.dsr = cnow.dsr;
	icount.rng = cnow.rng;
	icount.dcd = cnow.dcd;
	icount.rx = cnow.rx;
	icount.tx = cnow.tx;
	icount.frame = cnow.frame;
	icount.overrun = cnow.overrun;
	icount.parity = cnow.parity;
	icount.brk = cnow.brk;
	icount.buf_overrun = cnow.buf_overrun;

	return copy_to_user(user_icount, &icount, sizeof(icount)) ? -EFAULT : 0;
}

/*
 *  Checks for changes in the MSR lines CTS/DSR/RI/CD.  If change happened,
 *  wake up the sleeping process.  Both 1->0 and 0->1 transitions are counted except
 *  for RI where only 0->1 is counted.  RI not supported on all models, this driver is polled,
 *  so they are counts of MSR line changes, not actual interrupts.
 */
static void check_for_delta_msr(struct r_port *info)
{
	unsigned long new_msr;
	int wakeup_process = 0;
	static int first_time_through = 1;

	spin_lock_bh(&info->slock);

	if (!info->am_delta_msr_waiting) {
		spin_unlock_bh(&info->slock);
		return;
	}

	new_msr = get_msr(info);

	if (first_time_through) {
		first_time_through = 0;
		info->last_msr = new_msr;
	} else {
		if (new_msr != info->last_msr) {

			if ((new_msr & TIOCM_CTS) ^
			    (info->last_msr & TIOCM_CTS)) {
#ifdef ROCKET_DEBUG_DELTA_MSR
				printk(KERN_INFO
				       "Delta_msr() - CTS changed \n");
#endif
				info->icount.cts++;
				wakeup_process = 1;
			}

			if ((new_msr & TIOCM_DSR) ^
			    (info->last_msr & TIOCM_DSR)) {
#ifdef ROCKET_DEBUG_DELTA_MSR
				printk(KERN_INFO
				       "Delta_msr() - DSR changed \n");
#endif
				info->icount.dsr++;
				wakeup_process = 1;
			}

			if ((new_msr & TIOCM_CAR) ^
			    (info->last_msr & TIOCM_CAR)) {
#ifdef ROCKET_DEBUG_DELTA_MSR
				printk(KERN_INFO "Delta_msr() - CD changed \n");
#endif
				info->icount.dcd++;
				wakeup_process = 1;
			}

			/*  If RI 0->1  */
			if ((new_msr & TIOCM_RNG)
			    && (!(info->last_msr & TIOCM_RNG))) {
#ifdef ROCKET_DEBUG_DELTA_MSR
				printk(KERN_INFO
				       "Delta_msr() - RI went high \n");
#endif
				info->icount.rng++;
				wakeup_process = 1;
			}

			info->last_msr = new_msr;
		}
	}

	spin_unlock_bh(&info->slock);

	/*
	 *  We need the wakeup flag since the MSR will mismatch if RI goes 1->0.
	 *  We don't want to count this transition or wake up the sleeping process.
	 */
	if (wakeup_process)
		wake_up_interruptible(&info->delta_msr_wait);
}
#endif				/* no msr in 2.5.0+ */

/*  IOCTL call handler into the driver */
#if LINUX_VERSION_CODE < VERSION_CODE(2,6,39)
static int rp_ioctl(struct tty_struct *tty, struct file *file, unsigned int cmd, unsigned long arg)
#else
static int rp_ioctl(struct tty_struct *tty, unsigned int cmd, unsigned long arg)
#endif
{
	struct r_port *info = (struct r_port *)tty->driver_data;

	if (cmd != RCKP_GET_PORTS && rocket_paranoia_check(info, "rp_ioctl"))
		return -ENXIO;

	switch (cmd) {
	case TIOCMGET:
		return get_modem_info(info, arg);

	case TIOCMBIS:
	case TIOCMBIC:
	case TIOCMSET:
		return set_modem_info(info, cmd, arg);

#if LINUX_VERSION_CODE < VERSION_CODE(2,5,0)
	case TIOCMIWAIT:
		return wait_msr_change(info, (unsigned long *)arg);

	case TIOCGICOUNT:
		return get_delta_msr_counts(info,
					    (struct serial_icounter_struct *)
					    arg);
#endif
	case RCKP_GET_STRUCT:
		if (copy_to_user((void *)arg, info, sizeof(struct r_port)))
			return -EFAULT;
		return 0;
	case RCKP_GET_CONFIG:
		return get_config(info, (struct rocket_config *)arg);
	case RCKP_SET_CONFIG:
		return set_config(info, (struct rocket_config *)arg);
	case RCKP_GET_PORTS:
		return get_ports(info, (struct rocket_ports *)arg);
	case RCKP_RESET_RM2:
		return reset_rm2(info, arg);
	case RCKP_GET_VERSION:
		return get_version(info, (struct rocket_version *)arg);
	default:
		return -ENOIOCTLCMD;
	}
	return 0;
}

#if (defined(ROCKET_DEBUG_FLOW) || defined(ROCKET_DEBUG_THROTTLE))
static char *rp_tty_name(struct tty_struct *tty, char *buf)
{
	if (tty)
		sprintf(buf, "%s%d", TTY_DRIVER_NAME(tty),
			TTY_DRIVER_MINOR_START(tty) + TTY_DRIVER_NAME_BASE(tty));
	else
		strcpy(buf, "NULL tty");
	return buf;
}
#endif

static void rp_send_xchar(struct tty_struct *tty, char ch)
{
	struct r_port *info = (struct r_port *)tty->driver_data;
	CHANNEL_t *cp;

	if (rocket_paranoia_check(info, "rp_send_xchar"))
		return;

	cp = &info->channel;
	if (sGetTxCnt(cp))
		sWriteTxPrioByte(cp, ch);
	else
		sWriteTxByte(sGetTxRxDataIO(cp), ch);
}

static void rp_throttle(struct tty_struct *tty)
{
	struct r_port *info = (struct r_port *)tty->driver_data;
	CHANNEL_t *cp;

	if (rocket_paranoia_check(info, "rp_throttle"))
		return;

	cp = &info->channel;
	if (I_IXOFF(tty))
		rp_send_xchar(tty, STOP_CHAR(tty));

	sClrRTS(&info->channel);
}

static void rp_unthrottle(struct tty_struct *tty)
{
	struct r_port *info = (struct r_port *)tty->driver_data;
	CHANNEL_t *cp;

	if (rocket_paranoia_check(info, "rp_throttle"))
		return;

	cp = &info->channel;
	if (I_IXOFF(tty))
		rp_send_xchar(tty, START_CHAR(tty));

	sSetRTS(&info->channel);
}

/*
 * ------------------------------------------------------------
 * rp_stop() and rp_start()
 *
 * This routines are called before setting or resetting tty->stopped.
 * They enable or disable transmitter interrupts, as necessary.
 * ------------------------------------------------------------
 */
static void rp_stop(struct tty_struct *tty)
{
	struct r_port *info = (struct r_port *)tty->driver_data;
#ifdef ROCKET_DEBUG_FLOW
	char buf[64];
	printk(KERN_INFO "rp_stop: %s\n", rp_tty_name(tty, buf));
#endif

	if (rocket_paranoia_check(info, "rp_stop"))
		return;

	if (sGetTxCnt(&info->channel))
		sDisTransmit(&info->channel);
}

static void rp_start(struct tty_struct *tty)
{
	struct r_port *info = (struct r_port *)tty->driver_data;
#ifdef ROCKET_DEBUG_FLOW
	char buf[64];
	printk(KERN_INFO "rp_start: %s\n", rp_tty_name(tty, buf));
#endif

	if (rocket_paranoia_check(info, "rp_start"))
		return;

	sEnTransmit(&info->channel);
	set_bit((info->aiop * 8) + info->chan,
		(void *)&xmit_flags[info->board]);
}

/*
 * rp_wait_until_sent() --- wait until the transmitter is empty
 */
static void rp_wait_until_sent(struct tty_struct *tty, int timeout)
{
	struct r_port *info = (struct r_port *)tty->driver_data;
	CHANNEL_t *cp;
	unsigned long orig_jiffies;
	int check_time, exit_time;
	int txcnt;

	if (rocket_paranoia_check(info, "rp_wait_until_sent"))
		return;

	cp = &info->channel;

	orig_jiffies = jiffies;
#ifdef ROCKET_DEBUG_WAIT_UNTIL_SENT
	printk(KERN_INFO "In RP_wait_until_sent(%d) (jiff=%lu)...", timeout,
	       jiffies);
	printk(KERN_INFO "cps=%d...", info->cps);
#endif
	while (1) {
		txcnt = sGetTxCnt(cp);
		if (!txcnt) {
			if (sGetChanStatusLo(cp) & TXSHRMT)
				break;
			check_time = (HZ / info->cps) / 5;
		} else {
			check_time = HZ * txcnt / info->cps;
		}
		if (timeout) {
			// work-around for broken tty-layer code that passes
			// 64-bit value resulting in our seeing -1 for a
			// timeout instead of max signed value.
			if (timeout > 0)
				exit_time = orig_jiffies + timeout - jiffies;
			else
				exit_time = ((unsigned)~0)>>1;
			if (exit_time <= 0)
				break;
			if (exit_time < check_time)
				check_time = exit_time;
		}
		if (check_time == 0)
			check_time = 1;
#ifdef ROCKET_DEBUG_WAIT_UNTIL_SENT
		printk(KERN_INFO "txcnt = %d (jiff=%lu,check=%d)...", txcnt,
		       jiffies, check_time);
#endif
		current->state = TASK_INTERRUPTIBLE;
		schedule_timeout(check_time);
		if (signal_pending(current))
			break;
	}
	current->state = TASK_RUNNING;
#ifdef ROCKET_DEBUG_WAIT_UNTIL_SENT
	printk(KERN_INFO "txcnt = %d (jiff=%lu)...done\n", txcnt, jiffies);
#endif
}

/*
 * rp_hangup() --- called by tty_hangup() when a hangup is signaled.
 */
static void rp_hangup(struct tty_struct *tty)
{
	CHANNEL_t *cp;
	struct r_port *info = (struct r_port *)tty->driver_data;

	if (rocket_paranoia_check(info, "rp_hangup"))
		return;

#if (defined(ROCKET_DEBUG_OPEN) || defined(ROCKET_DEBUG_HANGUP))
	printk(KERN_INFO "rp_hangup of ttyR%d...", info->line);
#endif
	rp_flush_buffer(tty);
	if (info->flags & ROCKET_CLOSING)
		return;
	if (info->count) {
#if LINUX_VERSION_CODE < VERSION_CODE(2,5,0)
#if defined(MODULE)
		MOD_DEC_USE_COUNT;
#endif
#endif
		atomic_dec(&rp_num_ports_open);
	}

	clear_bit((info->aiop * 8) + info->chan, (void *)&xmit_flags[info->board]);

	info->count = 0;
	info->flags &= ~ROCKET_NORMAL_ACTIVE;
	info->tty = 0;

	cp = &info->channel;
	sDisRxFIFO(cp);
	sDisTransmit(cp);
	sDisInterrupts(cp, (TXINT_EN | MCINT_EN | RXINT_EN | SRCINT_EN | CHANINT_EN));
	sDisCTSFlowCtl(cp);
	sDisTxSoftFlowCtl(cp);
	sClrTxXOFF(cp);
	info->flags &= ~ROCKET_INITIALIZED;

	wake_up_interruptible(&info->open_wait);
#if LINUX_VERSION_CODE < VERSION_CODE(2,5,0)
	wake_up_interruptible(&info->delta_msr_wait);
#endif
}

/*
 *  Exception handler - write char routine.  The RocketPort driver uses a
 *  double-buffering strategy, with the twist that if the in-memory CPU
 *  buffer is empty, and there's space in the transmit FIFO, the
 *  writing routines will write directly to transmit FIFO.
 *  Write buffer and counters protected by spinlocks
 */
#if LINUX_VERSION_CODE < VERSION_CODE(2,6,26)
static void rp_put_char(struct tty_struct *tty, unsigned char ch)
#else
static int rp_put_char(struct tty_struct *tty, unsigned char ch)
#endif
{
	int ret = 0;
	CHANNEL_t *cp;
	struct r_port *info = (struct r_port *)tty->driver_data;

	if (rocket_paranoia_check(info, "rp_put_char")) {
#if LINUX_VERSION_CODE < VERSION_CODE(2,6,26)
		return;
#else
		return ret;
#endif
	}

#ifdef ROCKET_DEBUG_WRITE
	printk(KERN_INFO "rp_put_char %c...", ch);
#endif

	spin_lock_bh(&info->slock);
	cp = &info->channel;

	if (!tty->stopped && !tty->hw_stopped && info->xmit_fifo_room == 0)
		info->xmit_fifo_room = TXFIFO_SIZE - sGetTxCnt(cp);

	if (tty->stopped || tty->hw_stopped || info->xmit_fifo_room == 0 || info->xmit_cnt != 0) {
		info->xmit_buf[info->xmit_head++] = ch;
		info->xmit_head &= XMIT_BUF_SIZE - 1;
		info->xmit_cnt++;
		set_bit((info->aiop * 8) + info->chan,
			(void *)&xmit_flags[info->board]);
	} else {
		sOutB(sGetTxRxDataIO(cp), ch);
		info->xmit_fifo_room--;
		ret++;
	}
	spin_unlock_bh(&info->slock);

#if LINUX_VERSION_CODE > VERSION_CODE(2,6,25)
	return ret;
#endif
}

/*
 *  Exception handler - write routine, called when user app writes to the device.
 *  Spinlocks protect the info xmit members.
 *
 *  Our open source friends changed the tty layer in kernel version 2.6.10, removing
 *  the from_user parameter since the line discipline layer is required to pass data
 *  in a kernel space buffer.
 */
#if LINUX_VERSION_CODE < VERSION_CODE(2,6,10)
static int rp_write(struct tty_struct *tty, int from_user,
		    const unsigned char *buf, int count)
#else
static int rp_write(struct tty_struct *tty, const unsigned char *buf, int count)
#endif
{
	struct r_port *info = (struct r_port *)tty->driver_data;
	CHANNEL_t *cp;
	int c, retval = 0;

	if (count <= 0 || rocket_paranoia_check(info, "rp_write"))
		return 0;

#if LINUX_VERSION_CODE < VERSION_CODE(2,6,10)
	if (from_user) {
		count = MIN(count, PAGE_SIZE);
		if (down_interruptible(&info->tmp_buf_sem))
			return -ERESTARTSYS;
		if (copy_from_user(info->tmp_buf, buf, count)) {
			up(&info->tmp_buf_sem);
			return -EFAULT;
		}
		buf = info->tmp_buf;
	}
#endif

	spin_lock_bh(&info->slock);

#ifdef ROCKET_DEBUG_WRITE
	printk(KERN_INFO "rp_write %d chars \n", count);
#endif
	cp = &info->channel;

	if (!tty->stopped && !tty->hw_stopped && info->xmit_fifo_room < count)
		info->xmit_fifo_room = TXFIFO_SIZE - sGetTxCnt(cp);

	/* If the write queue for the port is empty, and there is FIFO space, stuff bytes. */
	if (!tty->stopped && !tty->hw_stopped && info->xmit_cnt == 0
	    && info->xmit_fifo_room > 0) {
		c = MIN(count, info->xmit_fifo_room);

		/*  Push data into FIFO, 2 bytes at a time */
		sOutStrW(sGetTxRxDataIO(cp), (unsigned short *)buf, c / 2);

		/*  If there is a byte remaining, write it */
		if (c & 1)
			sOutB(sGetTxRxDataIO(cp), buf[c - 1]);

		retval += c;
		buf += c;
		count -= c;
		info->xmit_fifo_room -= c;
	}

	/* If count is zero, we wrote it all and are done */
	if (!count)
		goto end;

	/*  Write remaining data into the port's xmit_buf */
	while (1) {
		if (info->tty == 0)	/*   Seemingly obligatory check... */
			goto end;

		c = MIN(count,
			MIN(XMIT_BUF_SIZE - info->xmit_cnt - 1,
			    XMIT_BUF_SIZE - info->xmit_head));
		if (c <= 0)
			break;

		memcpy(info->xmit_buf + info->xmit_head, buf, c);

		info->xmit_head = (info->xmit_head + c) & (XMIT_BUF_SIZE - 1);
		info->xmit_cnt += c;

		buf += c;
		count -= c;
		retval += c;
	}

	if ((retval > 0) && !tty->stopped && !tty->hw_stopped)
		set_bit((info->aiop * 8) + info->chan,
			(void *)&xmit_flags[info->board]);

end:

	spin_unlock_bh(&info->slock);
#if LINUX_VERSION_CODE < VERSION_CODE(2,6,10)
	if (from_user)
		up(&info->tmp_buf_sem);
#endif

	if (info->xmit_cnt < WAKEUP_CHARS) {
#ifdef TTYLAYER_NEW
		tty_wakeup(tty);
#else
		if ((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) && tty->ldisc.write_wakeup)
			(tty->ldisc.write_wakeup) (tty);
#endif
		wake_up_interruptible(&tty->write_wait);
#ifdef ROCKETPORT_HAVE_POLL_WAIT
		wake_up_interruptible(&tty->poll_wait);
#endif
	}

	return retval;
}

/*
 * Return the number of characters that can be sent.  We estimate
 * only using the in-memory transmit buffer only, and ignore the
 * potential space in the transmit FIFO.
 */
static int rp_write_room(struct tty_struct *tty)
{
	struct r_port *info = (struct r_port *)tty->driver_data;
	int ret;

	if (rocket_paranoia_check(info, "rp_write_room"))
		return 0;

	ret = XMIT_BUF_SIZE - info->xmit_cnt - 1;
	if (ret < 0)
		ret = 0;
#ifdef ROCKET_DEBUG_WRITE
	printk(KERN_INFO "rp_write_room returns %d...", ret);
#endif
	return ret;
}

/*
 * Return the number of characters in the buffer.  Again, this only
 * counts those characters in the in-memory transmit buffer.
 */
static int rp_chars_in_buffer(struct tty_struct *tty)
{
	struct r_port *info = (struct r_port *)tty->driver_data;
	CHANNEL_t *cp;

	if (rocket_paranoia_check(info, "rp_chars_in_buffer"))
		return 0;

	cp = &info->channel;

#ifdef ROCKET_DEBUG_WRITE
	printk(KERN_INFO "rp_chars_in_buffer returns %d...", info->xmit_cnt);
#endif
	return info->xmit_cnt;
}

/*
 *  Flushes the TX fifo for a port, deletes data in the xmit_buf stored in the
 *  r_port struct for the port.  Note that spinlock are used to protect info members,
 *  do not call this function if the spinlock is already held.
 */
static void rp_flush_buffer(struct tty_struct *tty)
{
	struct r_port *info = (struct r_port *)tty->driver_data;
	CHANNEL_t *cp;

	if (rocket_paranoia_check(info, "rp_flush_buffer"))
		return;

	spin_lock_bh(&info->slock);
	info->xmit_cnt = info->xmit_head = info->xmit_tail = 0;
	spin_unlock_bh(&info->slock);

	wake_up_interruptible(&tty->write_wait);
#ifdef ROCKETPORT_HAVE_POLL_WAIT
	wake_up_interruptible(&tty->poll_wait);
#endif

#ifdef TTYLAYER_NEW
	tty_wakeup(tty);
#else
	if ((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) && tty->ldisc.write_wakeup)
		(tty->ldisc.write_wakeup) (tty);
#endif

	cp = &info->channel;
	sFlushTxFIFO(cp);
}

#ifdef CONFIG_PCI

/*
 *  Called when a PCI card is found.  Retrieves and stores model information,
 *  init's aiopic and serial port hardware.
 *  Inputs:  cardnumber (0-n) and pci_dev struct from the kernel
 */
static int __init register_PCI(int cardnum, struct pci_dev *dev)
{
	int i, aiopnum, channelnum, ret;
	unsigned int class_rev;
	modelInfo_t *modelInfo;
	CONTROLLER_T *CtlP;

	if (!dev || pci_enable_device(dev))
		return 0;

	is_PCI[cardnum] = 1;
	CtlP = sCtlNumToCtlPtr(cardnum);
	CtlP->rocketModel.startingPortNumber = nextLineNumber;
	CtlP->CtlNum = cardnum;
	CtlP->rp_baud_base = DEFAULT_BAUD_BASE;

	modelInfo = GetModelInfo(dev->device);

	if (modelInfo == NULL)
		return 0;

	CtlP->rocketModel.model = modelInfo->model;
	CtlP->rocketModel.numPorts = modelInfo->numPorts;
	CtlP->NumAiop = modelInfo->numAiops;
	strcpy(CtlP->rocketModel.modelString, modelInfo->modelString);

	if (cardIsUPCI(CtlP)) {
		CtlP->BaseAiopicIOAddr = pci_resource_start(dev, 2);
		CtlP->BaseLocalConfigurationIOAddr = pci_resource_start(dev, 1);
		CtlP->IntStatusIOAddr =
			CtlP->BaseLocalConfigurationIOAddr + _PCI_9030_INT_CTRL;
		CtlP->AiopIntrBits = upci_aiop_intr_bits;
	} else {
		CtlP->BaseAiopicIOAddr = pci_resource_start(dev, 0);
		CtlP->IntStatusIOAddr = CtlP->BaseAiopicIOAddr + _PCI_INT_FUNC;
		CtlP->AiopIntrBits = aiop_intr_bits;
	}

	for (i = 0; i < 4; i++) {
		CtlP->AiopID[i] = AIOPID_NULL;
	}

	for (i = 0; i < CtlP->NumAiop; i++) {
		CtlP->AiopIO[i] = CtlP->BaseAiopicIOAddr + (i * 0x40);
		CtlP->AiopIntChanIO[i] = CtlP->AiopIO[i] + _INT_CHAN;
		CtlP->AiopID[i] = AIOPID_0001;
		CtlP->AiopNumChan[i] = modelInfo->numPorts / modelInfo->numAiops;

		if (cardIsRocketportPlus(CtlP)) {
			CtlP->clockPrescale = RP_PLUS_PRESCALE;
			SetClockPrescale(CtlP, i, RP_PLUS_PRESCALE);
			CtlP->rp_baud_base = 921600;
		} else {
			CtlP->clockPrescale = DEFAULT_PRESCALE;
			SetClockPrescale(CtlP, i, DEFAULT_PRESCALE);
			CtlP->rp_baud_base = 460800;
		}

#ifdef ROCKET_DEBUG_IO
		printk(KERN_INFO
		       "Register PCI debug AIOP %d - numPorts: %d  numAiops: %d \n",
		       (int)cardnum, (int)modelInfo->numPorts,
		       (int)modelInfo->numAiops);
		printk(KERN_INFO
		       "Register PCI debug AIOP %d - AiopIO: 0x%x AiopIntChanIO: 0x%x AiopID: 0x%x AiopNumChan: 0x%x \n",
		       i, CtlP->AiopIO[i], CtlP->AiopIntChanIO[i],
		       CtlP->AiopID[i], CtlP->AiopNumChan[i]);
#endif
	}

	sPCIControllerEOI(CtlP);	/* clear EOI if warm init */

	/*
	 *  The UPCI Quad/Octa cards use a single device ID (0x0805,0x080B,0x0812), but can
	 *  have either 4 or 8 ports, depending on the cable installed.  Bit 14 of offset 0x54
	 *  in the Local Configuration space determines which.  If set 8 port, else 4 port.
	 */
	if (CtlP->rocketModel.model == MODEL_UPCI_RP8OCTA ||
	    CtlP->rocketModel.model == MODEL_UPCI_RPP8 ||
	    CtlP->rocketModel.model == MODEL_UPCI_RPP8_422) {
		if (!
		    (sInW
		     (CtlP->BaseLocalConfigurationIOAddr +
		      _PCI_9030_GPIO_CTRL) & PCI_GPIO_CTRL_8PORT)) {
			if (CtlP->rocketModel.model == MODEL_UPCI_RP8OCTA) {
				strcpy(CtlP->rocketModel.modelString,
				       "RocketPort UPCI 4 port w/quad cable");
				CtlP->AiopNumChan[0] = 4;
				CtlP->rocketModel.numPorts = 4;
				CtlP->NumAiop = 1;
			}
			if (CtlP->rocketModel.model == MODEL_UPCI_RPP8) {
				strcpy(CtlP->rocketModel.modelString,
				       "RocketPort Plus UPCI 4 port w/quad cable");
				CtlP->AiopNumChan[0] = 4;
				CtlP->rocketModel.numPorts = 4;
				CtlP->NumAiop = 1;
			}
			if (CtlP->rocketModel.model == MODEL_UPCI_RPP8_422) {
				strcpy(CtlP->rocketModel.modelString,
				       "RocketPort Plus UPCI 4 port RS422");
				CtlP->AiopNumChan[0] = 4;
				CtlP->rocketModel.numPorts = 4;
				CtlP->NumAiop = 1;
			}
#ifdef ROCKET_DEBUG_IO
			printk
			("UPCI Quad/Octa, RP Plus Quad/Octa or RP Plus Quad/Octa RS422 Card detected a 4 port cable \n");
		} else {
			printk
			("UPCI Quad/Octa, RP Plus Quad/Octa or RP Plus Quad/Octa RS422 Card detected a 8 port cable \n");
#endif
		}
	}

	for (aiopnum = 0; aiopnum < CtlP->NumAiop; aiopnum++) {
		sResetAiopByNum(CtlP, aiopnum);
		for (channelnum = 0; channelnum < CtlP->AiopNumChan[aiopnum];
		     channelnum++)
			init_r_port(cardnum, aiopnum, channelnum, dev);
	}

	/*  Rocket modems must be reset */
	if (cardIsModem(CtlP)) {
#ifdef ROCKET_DEBUG_IO
		printk(KERN_INFO
		       "Comtrol PCI card is a modem, resetting modems and speaker\n");
#endif
		aiopnum = 0;	//if modem, maximum is 1 aiop
		for (channelnum = 0; channelnum < CtlP->AiopNumChan[aiopnum];
		     channelnum++)
			sPCIModemReset(CtlP, channelnum, 1);
		mdelay(500);
		for (channelnum = 0; channelnum < CtlP->AiopNumChan[aiopnum];
		     channelnum++)
			sPCIModemReset(CtlP, channelnum, 0);
		mdelay(500);
		rmSpeakerReset(CtlP, CtlP->rocketModel.model);
	}

	/*
	 *  PCI Rocketmodem 4/6 cards has 2 types, identified by the revision field in the
	 *  PCI configuration space.  0x01:  RocketModem II (must have firmware loaded).
	 *  0x02:  RMi (socketed, no load required).  We must set a flag so the daemon
	 *  knows whether to load firmware to this device.
	 */
	if ((CtlP->rocketModel.model == MODEL_RP6M)
	    || (CtlP->rocketModel.model == MODEL_RP4M)) {
		ret =
			pci_read_config_dword(dev, PCI_CLASS_REVISION, &class_rev);
		if (ret) {
			printk(KERN_INFO
			       "  Error during register_PCI(), unable to read config dword class_revision \n");
		} else {
			if ((class_rev & 0xFF) == 1)
				CtlP->rocketModel.loadrm2 = 1;
		}
	}

	printk(KERN_INFO
	       "Comtrol PCI controller #%d ID 0x%x found at address 0x%x - %d AIOP(s) (%s)\n",
	       cardnum, dev->device, CtlP->BaseAiopicIOAddr, CtlP->NumAiop,
	       CtlP->rocketModel.modelString);

#ifdef ROCKET_DEBUG_IO
	printk(KERN_INFO
	       "More card info:  AiopID[0]: %d, AiopNumChan[0]: %d AiopIntrBits: %x %x %x %x \n",
	       CtlP->AiopID[0], CtlP->AiopNumChan[0], CtlP->AiopIntrBits[0],
	       (CtlP->AiopIntrBits[1]), (CtlP->AiopIntrBits[2]),
	       (CtlP->AiopIntrBits[3]));
	printk(KERN_INFO
	       "More card info:  Local Configuration Addr: 0x%x IntStatus Addr: 0x%x   AIOP[0] IO Addr: 0x%x Aiop[0] Int Chan IO Addr:0x%x \n",
	       CtlP->BaseLocalConfigurationIOAddr, CtlP->IntStatusIOAddr,
	       CtlP->AiopIO[0], CtlP->AiopIntChanIO[0]);
#endif

	totalPorts += CtlP->rocketModel.numPorts;
	return (1);
}

/*
 *  Probes for PCI cards, inits them if found
 *  Input:   board_found = number of ISA boards already found, or the
 *           starting board number
 *  Returns: Number of PCI boards found
 */
static int __init init_PCI(int boards_found)
{
	struct pci_dev *dev = NULL;
	int count = 0;

	/*  Work through the PCI device list, pulling out ours */
	while ((dev = pci_get_device(PCI_VENDOR_ID_RP, PCI_ANY_ID, dev))) {
		if (register_PCI(count + boards_found, dev))
			count++;
	}
	return (count);
}

#endif				/* CONFIG_PCI */

/*
 *  Gets a region of ISA IO space.  This function is needed to keep the
 *  kernel version differences in one place.
 *  Inputs:  starting address and length to get.
 *  Returns:  0 if OK, -1 else
 */
static int get_isa_region(unsigned long start, unsigned long n)
{
	if (!request_region(start, n, "Comtrol RocketPort"))
		return -1;
	else
		return 0;
}

/*
 *  Probes for ISA cards
 *  Input:   i = the board number to look for
 *  Returns: 1 if board found, 0 else
 */
static int __init init_ISA(int i)
{
	int num_aiops, num_chan = 0, total_num_chan = 0;
	int aiop, chan;
	unsigned int aiopio[MAX_AIOPS_PER_BOARD];
	CONTROLLER_t *CtlP;
	char *type_string;

	//  If io_addr is zero, no board configured
	if (isa_io_addr[i] == 0)
		return (0);

	//  Reserve the IO region
	if (get_isa_region(isa_io_addr[i], 64)) {
		printk(KERN_INFO
		       "Unable to reserve IO region for configured ISA RocketPort at address 0x%lx, board not installed...\n",
		       isa_io_addr[i]);
		isa_io_addr[i] = 0;
		return (0);
	}

	CtlP = sCtlNumToCtlPtr(i);

	switch (isa_type[i]) {
	case ROCKET_TYPE_PC104:
		type_string = "(PC104)";
		break;
	case ROCKET_TYPE_MODEM:
		type_string = "(RocketModem)";
		break;
	default:
		type_string = "";
		break;
	}

	/*
	 * If support_low_speed is set, use the slow clock prescale,
	 * which supports 50 bps    if (reserved_controller == 0)
	 * request_region(controller, 4, "Comtrol Rocketport");
	 */
	if (support_low_speed) {
		CtlP->clockPrescale = 0x19;	/* mod 9 (divide by 10) prescale */
		CtlP->rp_baud_base = 230400;
	} else {
		CtlP->clockPrescale = 0x14;	/* mod 4 (devide by 5) prescale */
		CtlP->rp_baud_base = 460800;
	}

	for (aiop = 0; aiop < MAX_AIOPS_PER_BOARD; aiop++)
		aiopio[aiop] = isa_io_addr[i] + (aiop * 0x400);

	num_aiops =
		sInitController(CtlP, i, controller + (i * 0x400), aiopio,
				MAX_AIOPS_PER_BOARD, 0, FREQ_DIS, 0);
	if (cardIsPC104(CtlP)) {
		sEnAiop(CtlP, 2);	/* only one AIOPIC, but these */
		sEnAiop(CtlP, 3);	/* CSels used for other stuff */
	}
	//  If something went wrong initing the AIOP's release the ISA IO memory
	if (num_aiops <= 0) {
		release_region(isa_io_addr[i], 64);
		isa_io_addr[i] = 0;
		return (0);
	}

	CtlP->rocketModel.startingPortNumber = nextLineNumber;

	for (aiop = 0; aiop < num_aiops; aiop++) {
		sResetAiopByNum(CtlP, aiop);
		sEnAiop(CtlP, aiop);
		num_chan = sGetAiopNumChan(CtlP, aiop);
		total_num_chan += num_chan;
		for (chan = 0; chan < num_chan; chan++)
			init_r_port(i, aiop, chan, NULL);
	}
	is_PCI[i] = 0;
	if ((isa_type[i] == ROCKET_TYPE_MODEM)
	    || (isa_type[i] == ROCKET_TYPE_MODEMII)) {
		num_chan = sGetAiopNumChan(CtlP, 0);
		total_num_chan = num_chan;
		for (chan = 0; chan < num_chan; chan++)
			sModemReset(CtlP, chan, 1);
		mdelay(500);
		for (chan = 0; chan < num_chan; chan++)
			sModemReset(CtlP, chan, 0);
		mdelay(500);
		strcpy(CtlP->rocketModel.modelString, "RocketModem ISA");
	} else {
		strcpy(CtlP->rocketModel.modelString, "RocketPort ISA");
	}

	CtlP->rocketModel.numPorts = total_num_chan;
	CtlP->rocketModel.model = MODEL_ISA;

	printk(KERN_INFO
	       "RocketPort ISA card #%d found at 0x%lx - %d AIOPs %s\n", i,
	       isa_io_addr[i], num_aiops, type_string);
	printk(KERN_INFO "Installing %s, creating /dev/ttyR%d - %d\n",
	       CtlP->rocketModel.modelString,
	       CtlP->rocketModel.startingPortNumber,
	       CtlP->rocketModel.startingPortNumber +
	       CtlP->rocketModel.numPorts - 1);

	return (1);
}

#if LINUX_VERSION_CODE > VERSION_CODE(2,4,99)
static struct tty_operations rocket_ops = {
	.open = rp_open,
	.close = rp_close,
	.write = rp_write,
	.put_char = rp_put_char,
	.write_room = rp_write_room,
	.chars_in_buffer = rp_chars_in_buffer,
	.flush_buffer = rp_flush_buffer,
	.ioctl = rp_ioctl,
	.throttle = rp_throttle,
	.unthrottle = rp_unthrottle,
	.set_termios = rp_set_termios,
	.stop = rp_stop,
	.start = rp_start,
	.hangup = rp_hangup,
	.break_ctl = rp_break,
	.send_xchar = rp_send_xchar,
	.wait_until_sent = rp_wait_until_sent,
	.tiocmget = rp_tiocmget,
	.tiocmset = rp_tiocmset,
};
#endif

#ifdef CONFIG_SMP
#define SMPSUPPORTED "yes"
#else
#define SMPSUPPORTED "no"
#endif

/*
 * The module "startup" routine; it's run when the module is loaded.
 */
#if LINUX_VERSION_CODE < VERSION_CODE(2,5,0)
int __init init_module(void)
#else
static int __init rp_init(void)
#endif
{
	int retval, pci_boards_found, isa_boards_found, i;

	printk(KERN_INFO "RocketPort device driver module (SMP:%s), version %s, %s \n", SMPSUPPORTED, ROCKET_VERSION, ROCKET_DATE);

	// Initialize the array of pointers to our own internal state structures.
	memset(rp_table, 0, sizeof(rp_table));
	memset(xmit_flags, 0, sizeof(xmit_flags));

	for (i = 0; i < MAX_PORTS; i++)
		lineNumbers[i] = 0;
	nextLineNumber = 0;

	// If board 1 is non-zero, there is at least one ISA configured.  If
	//  controller is zero, use the default controller IO address of
	//  board1 + 0x40.

	if (board1) {
		if (controller == 0)
			controller = board1 + 0x40;
	} else {
		controller = 0;	//  Used as a flag, meaning no ISA boards */
	}

	//  If an ISA card is configured, reserve the 4 byte IO space for the Mudbac controller
	if (controller && (get_isa_region(controller, 4))) {
		printk(KERN_INFO
		       "Unable to reserve IO region for first configured ISA RocketPort controller 0x%lx.  Driver exiting \n",
		       controller);
		return -EBUSY;
	}

	//  Store ISA variable retrieved from command line or .conf file.
	isa_io_addr[0] = board1;
	isa_io_addr[1] = board2;
	isa_io_addr[2] = board3;
	isa_io_addr[3] = board4;

	isa_type[0] = modem1 ? ROCKET_TYPE_MODEM : ROCKET_TYPE_NORMAL;
	isa_type[0] = pc104_1[0] ? ROCKET_TYPE_PC104 : isa_type[0];
	isa_type[1] = modem2 ? ROCKET_TYPE_MODEM : ROCKET_TYPE_NORMAL;
	isa_type[1] = pc104_2[0] ? ROCKET_TYPE_PC104 : isa_type[1];
	isa_type[2] = modem3 ? ROCKET_TYPE_MODEM : ROCKET_TYPE_NORMAL;
	isa_type[2] = pc104_3[0] ? ROCKET_TYPE_PC104 : isa_type[2];
	isa_type[3] = modem4 ? ROCKET_TYPE_MODEM : ROCKET_TYPE_NORMAL;
	isa_type[3] = pc104_4[0] ? ROCKET_TYPE_PC104 : isa_type[3];

	// Set up the tty driver structure and then register this driver
	// with the tty layer.

	rocket_driver = alloc_tty_driver(MAX_PORTS);

// set flag that says we'll create individual devices later
#if defined(TTY_DRIVER_NO_DEVFS)
	rocket_driver->flags |= TTY_DRIVER_NO_DEVFS;
#endif
#if defined(TTY_DRIVER_DYNAMIC_DEV)
	rocket_driver->flags |= TTY_DRIVER_DYNAMIC_DEV;
#endif

	rocket_driver->name = "ttyR";
	rocket_driver->driver_name = "Comtrol RocketPort";
	rocket_driver->major = TTY_ROCKET_MAJOR;
	rocket_driver->minor_start = 0;
	rocket_driver->type = TTY_DRIVER_TYPE_SERIAL;
	rocket_driver->subtype = SERIAL_TYPE_NORMAL;
	rocket_driver->init_termios = tty_std_termios;
	rocket_driver->init_termios.c_cflag = B9600 | CS8 | CREAD | HUPCL | CLOCAL;

#if LINUX_VERSION_CODE > VERSION_CODE(2,6,20)
	rocket_driver->init_termios.c_ispeed = 9600;
	rocket_driver->init_termios.c_ospeed = 9600;
#endif

#if LINUX_VERSION_CODE > VERSION_CODE(2,5,0)
#if   LINUX_VERSION_CODE < VERSION_CODE(2,6,18)
	rocket_driver->devfs_name = "tts/R";	//ref http://www.tldp.org/HOWTO/Modem-HOWTO-9.html if needed.
#endif
	rocket_driver->owner = THIS_MODULE;

#if LINUX_VERSION_CODE < VERSION_CODE(2,6,26)
	rocket_driver->tiocmget = rp_tiocmget;
	rocket_driver->tiocmset = rp_tiocmset;
#endif
#endif
#if defined(ROCKET_SOFT_FLOW)
	rocket_driver->flags |= TTY_DRIVER_REAL_RAW;
#endif
#if LINUX_VERSION_CODE < VERSION_CODE(2,5,0)
	rocket_driver->refcount = &rocket_refcount;
	rocket_driver->table = rocket_table;
	rocket_driver->termios = rocket_termios;
	rocket_driver->termios_locked = rocket_termios_locked;

	rocket_driver->open = rp_open;
	rocket_driver->close = rp_close;
	rocket_driver->write = rp_write;
	rocket_driver->put_char = rp_put_char;
	rocket_driver->write_room = rp_write_room;
	rocket_driver->chars_in_buffer = rp_chars_in_buffer;
	rocket_driver->flush_buffer = rp_flush_buffer;
	rocket_driver->ioctl = rp_ioctl;
	rocket_driver->throttle = rp_throttle;
	rocket_driver->unthrottle = rp_unthrottle;
	rocket_driver->set_termios = rp_set_termios;
	rocket_driver->stop = rp_stop;
	rocket_driver->start = rp_start;
	rocket_driver->hangup = rp_hangup;
	rocket_driver->break_ctl = rp_break;
	rocket_driver->send_xchar = rp_send_xchar;
	rocket_driver->wait_until_sent = rp_wait_until_sent;
#else
	tty_set_operations(rocket_driver, &rocket_ops);
#endif

	retval = tty_register_driver(rocket_driver);
	if (retval < 0) {
		put_tty_driver(rocket_driver);
		printk(KERN_INFO "Couldn't install tty RocketPort driver (error %d)\n", -retval);
		return -1;
	}

#ifdef ROCKET_DEBUG_OPEN
	printk(KERN_INFO "RocketPort driver is major %d\n", rocket_driver->major);
#endif

	/*
	 *  OK, let's probe each of the controllers looking for boards.  Any boards found
	 *  will be initialized here.
	 */
	isa_boards_found = 0;
	pci_boards_found = 0;

	for (i = 0; i < MAX_NUMBEROF_BOARDS; i++) {
		if (init_ISA(i))
			isa_boards_found++;
	}

#ifdef CONFIG_PCI
	if (isa_boards_found < MAX_NUMBEROF_BOARDS)
		pci_boards_found = init_PCI(isa_boards_found);
#endif

	numberof_boards = pci_boards_found + isa_boards_found;

	if (numberof_boards == 0) {
		printk(KERN_INFO "No RocketPort ports found; unloading driver.\n");
		if (tty_unregister_driver(rocket_driver))
			printk(KERN_INFO "Error while trying to unregister RocketPort driver\n");
		put_tty_driver(rocket_driver);
		return -ENXIO;
	}

	// Set up the timer channel.
	init_timer(&rocket_timer);
	rocket_timer.function = rp_do_poll;

	return 0;
}

#ifdef MODULE

#if LINUX_VERSION_CODE < VERSION_CODE(2,5,0)
void cleanup_module(void)
#else
static void rp_cleanup_module(void)
#endif
{
	int retval;
	int i;

	del_timer_sync(&rocket_timer);

#if LINUX_VERSION_CODE > VERSION_CODE(2,5,0)
	for (i=0; i < totalPorts; i++)
		tty_unregister_device(rocket_driver, i);
#endif

	retval = tty_unregister_driver(rocket_driver);
	if (retval)
		printk(KERN_INFO "Error %d while trying to unregister RocketPort driver\n", -retval);
	put_tty_driver(rocket_driver);

	for (i = 0; i < MAX_PORTS; i++) {
		if (rp_table[i])
			kfree(rp_table[i]);
	}

	for (i = 0; i < MAX_NUMBEROF_BOARDS; i++) {
		if (isa_io_addr[i] <= 0 || is_PCI[i])
			continue;
		release_region(isa_io_addr[i], 64);
	}
	if (controller)
		release_region(controller, 4);
	printk(KERN_INFO "Unloading RocketPort driver: done\n");

}

#endif

/***********************************************************************
  Rocketport Accessor Functions

***********************************************************************/

/***************************************************************************
Function: sInitController
Purpose:  Initialization of controller global registers and controller
          structure.
Call:     sInitController(CtlP,CtlNum,MudbacIO,AiopIOList,AiopIOListSize,
                          IRQNum,Frequency,PeriodicOnly)
          CONTROLLER_T *CtlP; Ptr to controller structure
          int CtlNum; Controller number
          u32 MudbacIO; Mudbac base I/O address.
          u32 *AiopIOList; List of I/O addresses for each AIOP.
             This list must be in the order the AIOPs will be found on the
             controller.  Once an AIOP in the list is not found, it is
             assumed that there are no more AIOPs on the controller.
          int AiopIOListSize; Number of addresses in AiopIOList
          int IRQNum; Interrupt Request number.  Can be any of the following:
                         0: Disable global interrupts
                         3: IRQ 3
                         4: IRQ 4
                         5: IRQ 5
                         9: IRQ 9
                         10: IRQ 10
                         11: IRQ 11
                         12: IRQ 12
                         15: IRQ 15
          u8 Frequency: A flag identifying the frequency
                   of the periodic interrupt, can be any one of the following:
                      FREQ_DIS - periodic interrupt disabled
                      FREQ_137HZ - 137 Hertz
                      FREQ_69HZ - 69 Hertz
                      FREQ_34HZ - 34 Hertz
                      FREQ_17HZ - 17 Hertz
                      FREQ_9HZ - 9 Hertz
                      FREQ_4HZ - 4 Hertz
                   If IRQNum is set to 0 the Frequency parameter is
                   overidden, it is forced to a value of FREQ_DIS.
          int PeriodicOnly: TRUE if all interrupts except the periodic
                               interrupt are to be blocked.
                            FALSE is both the periodic interrupt and
                               other channel interrupts are allowed.
                            If IRQNum is set to 0 the PeriodicOnly parameter is
                               overidden, it is forced to a value of FALSE.
Return:   int: Number of AIOPs on the controller, or -1 if controller
               initialization failed.

Comments:
          If periodic interrupts are to be disabled but AIOP interrupts
          are allowed, set Frequency to FREQ_DIS and PeriodicOnly to FALSE.

          If interrupts are to be completely disabled set IRQNum to 0.

          Setting Frequency to FREQ_DIS and PeriodicOnly to TRUE is an
          invalid combination.

          This function performs initialization of global interrupt modes,
          but it does not actually enable global interrupts.  To enable
          and disable global interrupts use functions sEnGlobalInt() and
          sDisGlobalInt().  Enabling of global interrupts is normally not
          done until all other initializations are complete.

          Even if interrupts are globally enabled, they must also be
          individually enabled for each channel that is to generate
          interrupts.

Warnings: No range checking on any of the parameters is done.

          No context switches are allowed while executing this function.

          After this function all AIOPs on the controller are disabled,
          they can be enabled with sEnAiop().
*/
/*
static int sInitController(CONTROLLER_T * CtlP, int CtlNum, ByteIO_t MudbacIO,
			   ByteIO_t * AiopIOList, int AiopIOListSize,
			   int IRQNum, Byte_t Frequency, int PeriodicOnly)
*/
static int sInitController(CONTROLLER_T * CtlP, int CtlNum, u32 MudbacIO,
			   u32 * AiopIOList, int AiopIOListSize,
			   int IRQNum, u8 Frequency, int PeriodicOnly)
{
	int i;
	u32 io;
	int done;

	CtlP->AiopIntrBits = aiop_intr_bits;
	CtlP->CtlNum = CtlNum;
	CtlP->MBaseIO = MudbacIO;
	CtlP->MReg1IO = MudbacIO + 1;
	CtlP->MReg2IO = MudbacIO + 2;
	CtlP->MReg3IO = MudbacIO + 3;
#if 1
	CtlP->MReg2 = 0;	/* interrupt disable */
	CtlP->MReg3 = 0;	/* no periodic interrupts */
#else
	if (sIRQMap[IRQNum] == 0) {	/* interrupts globally disabled */
		CtlP->MReg2 = 0;	/* interrupt disable */
		CtlP->MReg3 = 0;	/* no periodic interrupts */
	} else {
		CtlP->MReg2 = sIRQMap[IRQNum];	/* set IRQ number */
		CtlP->MReg3 = Frequency;	/* set frequency */
		if (PeriodicOnly) {	/* periodic interrupt only */
			CtlP->MReg3 |= PERIODIC_ONLY;
		}
	}
#endif
	sOutB(CtlP->MReg2IO, CtlP->MReg2);
	sOutB(CtlP->MReg3IO, CtlP->MReg3);
	sControllerEOI(CtlP);	/* clear EOI if warm init */
	/* Init AIOPs */
	CtlP->NumAiop = 0;
	for (i = done = 0; i < AiopIOListSize; i++) {
		io = AiopIOList[i];
		CtlP->AiopIO[i] = (u32) io;
		CtlP->AiopIntChanIO[i] = io + _INT_CHAN;
		sOutB(CtlP->MReg2IO, CtlP->MReg2 | (i & 0x03));	/* AIOP index */
		sOutB(MudbacIO, (u8) (io >> 6));	/* set up AIOP I/O in MUDBAC */
		if (done)
			continue;
		sEnAiop(CtlP, i);	/* enable the AIOP */
		CtlP->AiopID[i] = sReadAiopID(io);	/* read AIOP ID */
		if (CtlP->AiopID[i] == AIOPID_NULL)	/* if AIOP does not exist */
			done = 1;	/* done looking for AIOPs */
		else {
			CtlP->AiopNumChan[i] = sReadAiopNumChan((u32) io);	/* num channels in AIOP */
			sOutW((u32) io + _INDX_ADDR, _CLK_PRE);	/* clock prescaler */
			sOutB(io + _INDX_DATA, CtlP->clockPrescale);
			CtlP->NumAiop++;	/* bump count of AIOPs */
		}
		sDisAiop(CtlP, i);	/* disable AIOP */
	}

	if (CtlP->NumAiop == 0)
		return (-1);
	else
		return (CtlP->NumAiop);
}

/***************************************************************************
Function: sReadAiopID
Purpose:  Read the AIOP idenfication number directly from an AIOP.
Call:     sReadAiopID(io)
          u32 io: AIOP base I/O address
Return:   int: Flag AIOPID_XXXX if a valid AIOP is found, where X
                 is replace by an identifying number.
          Flag AIOPID_NULL if no valid AIOP is found
Warnings: No context switches are allowed while executing this function.

*/
static int sReadAiopID(u32 io)
{
	u8 AiopID;		/* ID byte from AIOP */
	sOutB(io + _CMD_REG, RESET_ALL);	/* reset AIOP */
	sOutB(io + _CMD_REG, 0x0);
	AiopID = sInW(io + _CHN_STAT0) & 0x07;
	if (AiopID == 0x06)
		return (1);
	else			/* AIOP does not exist */
		return (-1);
}

/***************************************************************************
Function: sReadAiopNumChan
Purpose:  Read the number of channels available in an AIOP directly from
          an AIOP.
Call:     sReadAiopNumChan(io)
          u32 io: AIOP base I/O address
Return:   int: The number of channels available
Comments: The number of channels is determined by write/reads from identical
          offsets within the SRAM address spaces for channels 0 and 4.
          If the channel 4 space is mirrored to channel 0 it is a 4 channel
          AIOP, otherwise it is an 8 channel.
Warnings: No context switches are allowed while executing this function.
*/
static int sReadAiopNumChan(u32 io)
{
	u32 x;
	static u8 R[4] = { 0x00, 0x00, 0x34, 0x12 };

	/* write to chan 0 SRAM */
	sOutDW((u32) io + _INDX_ADDR, *((u32 *) & R[0]));
	sOutW(io + _INDX_ADDR, 0);	/* read from SRAM, chan 0 */
	x = sInW(io + _INDX_DATA);
	sOutW(io + _INDX_ADDR, 0x4000);	/* read from SRAM, chan 4 */
	if (x != sInW(io + _INDX_DATA))	/* if different must be 8 chan */
		return (8);
	else
		return (4);
}

/***************************************************************************
Function: sInitChan
Purpose:  Initialization of a channel and channel structure
Call:     sInitChan(CtlP,ChP,AiopNum,ChanNum)
          CONTROLLER_T *CtlP; Ptr to controller structure
          CHANNEL_T *ChP; Ptr to channel structure
          int AiopNum; AIOP number within controller
          int ChanNum; Channel number within AIOP
Return:   int: TRUE if initialization succeeded, FALSE if it fails because channel
               number exceeds number of channels available in AIOP.
Comments: This function must be called before a channel can be used.
Warnings: No range checking on any of the parameters is done.

          No context switches are allowed while executing this function.
*/
static int sInitChan(CONTROLLER_T * CtlP, CHANNEL_T * ChP, int AiopNum, int ChanNum)
{
	int i;
	u32 AiopIO;
	u32 ChIOOff;
	u8 *ChR;
	u32 ChOff;
	static u8 R[4];
	int brd9600;

	if (ChanNum >= CtlP->AiopNumChan[AiopNum])
		return (FALSE);	/* exceeds num chans in AIOP */

	/* Channel, AIOP, and controller identifiers */
	ChP->CtlP = CtlP;
	ChP->ChanID = CtlP->AiopID[AiopNum];
	ChP->AiopNum = AiopNum;
	ChP->ChanNum = ChanNum;

	/* Global direct addresses */
	AiopIO = CtlP->AiopIO[AiopNum];
	ChP->Cmd = (u32) AiopIO + _CMD_REG;
	ChP->IntChan = (u32) AiopIO + _INT_CHAN;
	ChP->IntMask = (u32) AiopIO + _INT_MASK;
	ChP->IndexAddr = (u32) AiopIO + _INDX_ADDR;
	ChP->IndexData = AiopIO + _INDX_DATA;

	/* Channel direct addresses */
	ChIOOff = AiopIO + ChP->ChanNum * 2;
	ChP->TxRxData = ChIOOff + _TD0;
	ChP->ChanStat = ChIOOff + _CHN_STAT0;
	ChP->TxRxCount = ChIOOff + _FIFO_CNT0;
	ChP->IntID = (u32) AiopIO + ChP->ChanNum + _INT_ID0;

	/* Initialize the channel from the RData array */
	for (i = 0; i < RDATASIZE; i += 4) {
		R[0] = RData[i];
		R[1] = RData[i + 1] + 0x10 * ChanNum;
		R[2] = RData[i + 2];
		R[3] = RData[i + 3];
		sOutDW(ChP->IndexAddr, *((u32 *) & R[0]));
	}

	ChR = ChP->R;
	for (i = 0; i < RREGDATASIZE; i += 4) {
		ChR[i] = RRegData[i];
		ChR[i + 1] = RRegData[i + 1] + 0x10 * ChanNum;
		ChR[i + 2] = RRegData[i + 2];
		ChR[i + 3] = RRegData[i + 3];
	}

	/* Indexed registers */
	ChOff = (u32) ChanNum *0x1000;

	if (ChP->CtlP->clockPrescale == 0x14)
		brd9600 = 47;
	else
		brd9600 = 23;
	ChP->BaudDiv[0] = (u8) (ChOff + _BAUD);
	ChP->BaudDiv[1] = (u8) ((ChOff + _BAUD) >> 8);
	ChP->BaudDiv[2] = (u8) brd9600;
	ChP->BaudDiv[3] = (u8) (brd9600 >> 8);
	sOutDW(ChP->IndexAddr, *(u32 *) & ChP->BaudDiv[0]);

	ChP->TxControl[0] = (u8) (ChOff + _TX_CTRL);
	ChP->TxControl[1] = (u8) ((ChOff + _TX_CTRL) >> 8);
	ChP->TxControl[2] = 0;
	ChP->TxControl[3] = 0;
	sOutDW(ChP->IndexAddr, *(u32 *) & ChP->TxControl[0]);

	ChP->RxControl[0] = (u8) (ChOff + _RX_CTRL);
	ChP->RxControl[1] = (u8) ((ChOff + _RX_CTRL) >> 8);
	ChP->RxControl[2] = 0;
	ChP->RxControl[3] = 0;
	sOutDW(ChP->IndexAddr, *(u32 *) & ChP->RxControl[0]);

	ChP->TxEnables[0] = (u8) (ChOff + _TX_ENBLS);
	ChP->TxEnables[1] = (u8) ((ChOff + _TX_ENBLS) >> 8);
	ChP->TxEnables[2] = 0;
	ChP->TxEnables[3] = 0;
	sOutDW(ChP->IndexAddr, *(u32 *) & ChP->TxEnables[0]);

	ChP->TxCompare[0] = (u8) (ChOff + _TXCMP1);
	ChP->TxCompare[1] = (u8) ((ChOff + _TXCMP1) >> 8);
	ChP->TxCompare[2] = 0;
	ChP->TxCompare[3] = 0;
	sOutDW(ChP->IndexAddr, *(u32 *) & ChP->TxCompare[0]);

	ChP->TxReplace1[0] = (u8) (ChOff + _TXREP1B1);
	ChP->TxReplace1[1] = (u8) ((ChOff + _TXREP1B1) >> 8);
	ChP->TxReplace1[2] = 0;
	ChP->TxReplace1[3] = 0;
	sOutDW(ChP->IndexAddr, *(u32 *) & ChP->TxReplace1[0]);

	ChP->TxReplace2[0] = (u8) (ChOff + _TXREP2);
	ChP->TxReplace2[1] = (u8) ((ChOff + _TXREP2) >> 8);
	ChP->TxReplace2[2] = 0;
	ChP->TxReplace2[3] = 0;
	sOutDW(ChP->IndexAddr, *(u32 *) & ChP->TxReplace2[0]);

	ChP->TxFIFOPtrs = ChOff + _TXF_OUTP;
	ChP->TxFIFO = ChOff + _TX_FIFO;

	sOutB(ChP->Cmd, (u8) ChanNum | RESTXFCNT);	// apply reset Tx FIFO count
	sOutB(ChP->Cmd, (u8) ChanNum);	// remove reset Tx FIFO count
	sOutW((u32) ChP->IndexAddr, ChP->TxFIFOPtrs);	// clear Tx in/out ptrs
	sOutW(ChP->IndexData, 0);
	ChP->RxFIFOPtrs = ChOff + _RXF_OUTP;
	ChP->RxFIFO = ChOff + _RX_FIFO;

	sOutB(ChP->Cmd, (u8) ChanNum | RESRXFCNT);	// apply reset Rx FIFO count
	sOutB(ChP->Cmd, (u8) ChanNum);	// remove reset Rx FIFO count
	sOutW((u32) ChP->IndexAddr, ChP->RxFIFOPtrs);	// clear Rx out ptr
	sOutW(ChP->IndexData, 0);
	sOutW((u32) ChP->IndexAddr, ChP->RxFIFOPtrs + 2);	/* clear Rx in ptr */
	sOutW(ChP->IndexData, 0);
	ChP->TxPrioCnt = ChOff + _TXP_CNT;
	sOutW((u32) ChP->IndexAddr, ChP->TxPrioCnt);
	sOutB(ChP->IndexData, 0);
	ChP->TxPrioPtr = ChOff + _TXP_PNTR;
	sOutW((u32) ChP->IndexAddr, ChP->TxPrioPtr);
	sOutB(ChP->IndexData, 0);
	ChP->TxPrioBuf = ChOff + _TXP_BUF;
	sEnRxProcessor(ChP);	/* start the Rx processor */

	return (TRUE);
}

/***************************************************************************
Function: sStopRxProcessor
Purpose:  Stop the receive processor from processing a channel.
Call:     sStopRxProcessor(ChP)
          CHANNEL_T *ChP; Ptr to channel structure

Comments: The receive processor can be started again with sStartRxProcessor().
          This function causes the receive processor to skip over the
          stopped channel.  It does not stop it from processing other channels.

Warnings: No context switches are allowed while executing this function.

          Do not leave the receive processor stopped for more than one
          character time.

          After calling this function a delay of 4 uS is required to ensure
          that the receive processor is no longer processing this channel.
*/
static void sStopRxProcessor(CHANNEL_T * ChP)
{
	u8 R[4];
	R[0] = ChP->R[0];
	R[1] = ChP->R[1];
	R[2] = 0x0a;
	R[3] = ChP->R[3];
	sOutDW(ChP->IndexAddr, *(u32 *) & R[0]);
}

/***************************************************************************
Function: sFlushRxFIFO
Purpose:  Flush the Rx FIFO
Call:     sFlushRxFIFO(ChP)
          CHANNEL_T *ChP; Ptr to channel structure
Return:   void
Comments: To prevent data from being enqueued or dequeued in the Tx FIFO
          while it is being flushed the receive processor is stopped
          and the transmitter is disabled.  After these operations a
          4 uS delay is done before clearing the pointers to allow
          the receive processor to stop.  These items are handled inside
          this function.
Warnings: No context switches are allowed while executing this function.
*/
static void sFlushRxFIFO(CHANNEL_T * ChP)
{
	int i;
	u8 Ch;		// channel number within AIOP
	int RxFIFOEnabled;	/* TRUE if Rx FIFO enabled */

	if (sGetRxCnt(ChP) == 0)	/* Rx FIFO empty */
		return;		/* don't need to flush */

	RxFIFOEnabled = FALSE;
	if (ChP->R[0x32] == 0x08) {	/* Rx FIFO is enabled */
		RxFIFOEnabled = TRUE;
		sDisRxFIFO(ChP);	/* disable it */
		for (i = 0; i < 2000 / 200; i++)	/* delay 2 uS to allow proc to disable FIFO */
			sInB(ChP->IntChan);	/* depends on bus i/o timing */
	}
	sGetChanStatus(ChP);	/* clear any pending Rx errors in chan stat */
	Ch = (u8) sGetChanNum(ChP);
	sOutB(ChP->Cmd, Ch | RESRXFCNT);	/* apply reset Rx FIFO count */
	sOutB(ChP->Cmd, Ch);	/* remove reset Rx FIFO count */
	sOutW((u32) ChP->IndexAddr, ChP->RxFIFOPtrs);	/* clear Rx out ptr */
	sOutW(ChP->IndexData, 0);
	sOutW((u32) ChP->IndexAddr, ChP->RxFIFOPtrs + 2);	/* clear Rx in ptr */
	sOutW(ChP->IndexData, 0);
	if (RxFIFOEnabled)
		sEnRxFIFO(ChP);	/* enable Rx FIFO */
}

/***************************************************************************
Function: sFlushTxFIFO
Purpose:  Flush the Tx FIFO
Call:     sFlushTxFIFO(ChP)
          CHANNEL_T *ChP; Ptr to channel structure
Return:   void
Comments: To prevent data from being enqueued or dequeued in the Tx FIFO
          while it is being flushed the receive processor is stopped
          and the transmitter is disabled.  After these operations a
          4 uS delay is done before clearing the pointers to allow
          the receive processor to stop.  These items are handled inside
          this function.
Warnings: No context switches are allowed while executing this function.
*/
static void sFlushTxFIFO(CHANNEL_T * ChP)
{
	int i;
	u8 Ch;		// channel number within AIOP
	int TxEnabled;		/* TRUE if transmitter enabled */

	if (sGetTxCnt(ChP) == 0)	/* Tx FIFO empty */
		return;		/* don't need to flush */

	TxEnabled = FALSE;
	if (ChP->TxControl[3] & TX_ENABLE) {
		TxEnabled = TRUE;
		sDisTransmit(ChP);	/* disable transmitter */
	}
	sStopRxProcessor(ChP);	/* stop Rx processor */
	for (i = 0; i < 4000 / 200; i++)	/* delay 4 uS to allow proc to stop */
		sInB(ChP->IntChan);	/* depends on bus i/o timing */
	Ch = (u8) sGetChanNum(ChP);
	sOutB(ChP->Cmd, Ch | RESTXFCNT);	/* apply reset Tx FIFO count */
	sOutB(ChP->Cmd, Ch);	/* remove reset Tx FIFO count */
	sOutW((u32) ChP->IndexAddr, ChP->TxFIFOPtrs);	/* clear Tx in/out ptrs */
	sOutW(ChP->IndexData, 0);
	if (TxEnabled)
		sEnTransmit(ChP);	/* enable transmitter */
	sStartRxProcessor(ChP);	/* restart Rx processor */
}

/***************************************************************************
Function: sWriteTxPrioByte
Purpose:  Write a byte of priority transmit data to a channel
Call:     sWriteTxPrioByte(ChP,Data)
          CHANNEL_T *ChP; Ptr to channel structure
          u8 Data; The transmit data byte

Return:   int: 1 if the bytes is successfully written, otherwise 0.

Comments: The priority byte is transmitted before any data in the Tx FIFO.

Warnings: No context switches are allowed while executing this function.
*/
//static int sWriteTxPrioByte(CHANNEL_T * ChP, Byte_t Data)
static int sWriteTxPrioByte(CHANNEL_T * ChP, u8 Data)
{
	u8 DWBuf[4];	// buffer for double word writes
	u32 *WordPtr;	/* must be far because Win SS != DS */
	register u32 IndexAddr;	//do not use the address operator on this!

	if (sGetTxCnt(ChP) > 1) {	/* write it to Tx priority buffer */
		IndexAddr = ChP->IndexAddr;
		sOutW((u32) IndexAddr, ChP->TxPrioCnt);	/* get priority buffer status */
		if (sInB((u32) ChP->IndexData) & PRI_PEND)	/* priority buffer busy */
			return (0);	/* nothing sent */

		WordPtr = (u32 *) (&DWBuf[0]);
		*WordPtr = ChP->TxPrioBuf;	/* data byte address */

		DWBuf[2] = Data;	/* data byte value */
		sOutDW(IndexAddr, *((u32 *) (&DWBuf[0])));	/* write it out */

		*WordPtr = ChP->TxPrioCnt;	/* Tx priority count address */

		DWBuf[2] = PRI_PEND + 1;	/* indicate 1 byte pending */
		DWBuf[3] = 0;	/* priority buffer pointer */
		sOutDW(IndexAddr, *((u32 *) (&DWBuf[0])));	/* write it out */
	} else {		/* write it to Tx FIFO */

		sWriteTxByte(sGetTxRxDataIO(ChP), Data);
	}
	return (1);		/* 1 byte sent */
}

/***************************************************************************
Function: sEnInterrupts
Purpose:  Enable one or more interrupts for a channel
Call:     sEnInterrupts(ChP,Flags)
          CHANNEL_T *ChP; Ptr to channel structure
          u32 Flags: Interrupt enable flags, can be any combination
             of the following flags:
                TXINT_EN:   Interrupt on Tx FIFO empty
                RXINT_EN:   Interrupt on Rx FIFO at trigger level (see
                            sSetRxTrigger())
                SRCINT_EN:  Interrupt on SRC (Special Rx Condition)
                MCINT_EN:   Interrupt on modem input change
                CHANINT_EN: Allow channel interrupt signal to the AIOP's
                            Interrupt Channel Register.
Return:   void
Comments: If an interrupt enable flag is set in Flags, that interrupt will be
          enabled.  If an interrupt enable flag is not set in Flags, that
          interrupt will not be changed.  Interrupts can be disabled with
          function sDisInterrupts().

          This function sets the appropriate bit for the channel in the AIOP's
          Interrupt Mask Register if the CHANINT_EN flag is set.  This allows
          this channel's bit to be set in the AIOP's Interrupt Channel Register.

          Interrupts must also be globally enabled before channel interrupts
          will be passed on to the host.  This is done with function
          sEnGlobalInt().

          In some cases it may be desirable to disable interrupts globally but
          enable channel interrupts.  This would allow the global interrupt
          status register to be used to determine which AIOPs need service.
*/
static void sEnInterrupts(CHANNEL_T * ChP, u32 Flags)
{
	u8 Mask;		// Interrupt Mask Register

	ChP->RxControl[2] |= ((u8) Flags & (RXINT_EN | SRCINT_EN | MCINT_EN));

	sOutDW(ChP->IndexAddr, *(u32 *) & ChP->RxControl[0]);

	ChP->TxControl[2] |= ((u8) Flags & TXINT_EN);

	sOutDW(ChP->IndexAddr, *(u32 *) & ChP->TxControl[0]);

	if (Flags & CHANINT_EN) {
		Mask = sInB(ChP->IntMask) | sBitMapSetTbl[ChP->ChanNum];
		sOutB(ChP->IntMask, Mask);
	}
}

/***************************************************************************
Function: sDisInterrupts
Purpose:  Disable one or more interrupts for a channel
Call:     sDisInterrupts(ChP,Flags)
          CHANNEL_T *ChP; Ptr to channel structure
          u32 Flags: Interrupt flags, can be any combination
             of the following flags:
                TXINT_EN:   Interrupt on Tx FIFO empty
                RXINT_EN:   Interrupt on Rx FIFO at trigger level (see
                            sSetRxTrigger())
                SRCINT_EN:  Interrupt on SRC (Special Rx Condition)
                MCINT_EN:   Interrupt on modem input change
                CHANINT_EN: Disable channel interrupt signal to the
                            AIOP's Interrupt Channel Register.
Return:   void
Comments: If an interrupt flag is set in Flags, that interrupt will be
          disabled.  If an interrupt flag is not set in Flags, that
          interrupt will not be changed.  Interrupts can be enabled with
          function sEnInterrupts().

          This function clears the appropriate bit for the channel in the AIOP's
          Interrupt Mask Register if the CHANINT_EN flag is set.  This blocks
          this channel's bit from being set in the AIOP's Interrupt Channel
          Register.
*/
static void sDisInterrupts(CHANNEL_T * ChP, u32 Flags)
{
	u8 Mask;		// Interrupt Mask Register

	ChP->RxControl[2] &= ~((u8) Flags & (RXINT_EN | SRCINT_EN | MCINT_EN));
	sOutDW(ChP->IndexAddr, *(u32 *) & ChP->RxControl[0]);
	ChP->TxControl[2] &= ~((u8) Flags & TXINT_EN);
	sOutDW(ChP->IndexAddr, *(u32 *) & ChP->TxControl[0]);

	if (Flags & CHANINT_EN) {
		Mask = sInB(ChP->IntMask) & sBitMapClrTbl[ChP->ChanNum];
		sOutB(ChP->IntMask, Mask);
	}
}

//static void sSetInterfaceMode(CHANNEL_T * ChP, Byte_t mode)
static void sSetInterfaceMode(CHANNEL_T * ChP, u8 mode)
{
	sOutB(ChP->CtlP->AiopIO[2], (mode & 0x18) | ChP->ChanNum);
}

/*
 *  Not an official SSCI function, but how to reset RocketModems.
 *  ISA bus version
 */
static void sModemReset(CONTROLLER_T * CtlP, int chan, int on)
{
	u32 addr;
	u8 val;

	addr = CtlP->AiopIO[0] + 0x400;
	val = sInB(CtlP->MReg3IO);
	/* if AIOP[1] is not enabled, enable it */
	if ((val & 2) == 0) {
		val = sInB(CtlP->MReg2IO);
		sOutB(CtlP->MReg2IO, (val & 0xfc) | (1 & 0x03));
		sOutB(CtlP->MBaseIO, (unsigned char)(addr >> 6));
	}

	sEnAiop(CtlP, 1);
	if (!on)
		addr += 8;
	sOutB(addr + chan, 0);	/* apply or remove reset */
	sDisAiop(CtlP, 1);
}

/*
 *  Resets the modems on PCI cards.
 */
static void sPCIModemReset(CONTROLLER_T * CtlP, int chan, int on)
{
	u32 addr;

	addr = CtlP->AiopIO[0] + 0x40;	/* 2nd AIOP */
	if (!on)
		addr += 8;
	sOutB(addr + chan, 0);	/* apply or remove reset */
}

/*  Resets the speaker controller on RocketModem II and III devices */
static void rmSpeakerReset(CONTROLLER_T * CtlP, int model)
{
	u32 addr;

	/* RocketModem II speaker control is at the 8th port location of offset 0x40 */
	if ((model == MODEL_RP4M) || (model == MODEL_RP6M)) {
		addr = CtlP->AiopIO[0] + 0x4F;
		sOutB(addr, 0);
	}

	/* RocketModem III/IV speaker control is at the 1st port location of offset 0x80 */
	if ((model == MODEL_UPCI_RM3_8PORT) || (model == MODEL_UPCI_RM3_4PORT) ||
	    (model == MODEL_UPCI_RM4_8PORT) || (model == MODEL_UPCI_RM4_4PORT)) {
		addr = CtlP->AiopIO[0] + 0x88;
		sOutB(addr, 0);
	}
}

/*  Returns the line number given the controller (board), aiop and channel number */
static unsigned char GetLineNumber(int ctrl, int aiop, int ch)
{
	return lineNumbers[(ctrl << 5) | (aiop << 3) | ch];
}

/*
 *  Stores the line number associated with a given controller (board), aiop
 *  and channel number.
 *  Returns:  The line number assigned
 */
static unsigned char SetLineNumber(int ctrl, int aiop, int ch)
{
	lineNumbers[(ctrl << 5) | (aiop << 3) | ch] = nextLineNumber++;
	return (nextLineNumber - 1);
}

/*
 *  Functions to determine RocketPort model types.
 *  Inputs:  Controller struct pointer for the card.
 *  Returns: 1 if yes, else 0
 */
static int cardIsUPCI(CONTROLLER_T * CtlP)
{
	if (CtlP->rocketModel.model & 0x0800)
		return TRUE;
	else
		return FALSE;
}

static int cardIsPCI(CONTROLLER_T * CtlP)
{
	if (cardIsUPCI(CtlP))
		return TRUE;

	if (CtlP->rocketModel.model & 0x000F)
		return TRUE;
	else
		return FALSE;
}

static int cardIsModem(CONTROLLER_T * CtlP)
{
	if ((CtlP->rocketModel.model == MODEL_RP6M) ||
	    (CtlP->rocketModel.model == MODEL_RP4M) ||
	    (CtlP->rocketModel.model == MODEL_UPCI_RM3_8PORT) ||
	    (CtlP->rocketModel.model == MODEL_UPCI_RM3_4PORT) ||
	    (CtlP->rocketModel.model == MODEL_UPCI_RM4_8PORT) ||
	    (CtlP->rocketModel.model == MODEL_UPCI_RM4_4PORT))
		return TRUE;
	else
		return FALSE;
}


static int cardIsPC104(CONTROLLER_T * CtlP)
{
	if (isa_type[CtlP->CtlNum] == ROCKET_TYPE_PC104)
		return TRUE;
	else
		return FALSE;
}

static int cardIsRM2(CONTROLLER_T * CtlP)
{
	if ((CtlP->rocketModel.model == MODEL_RP6M) ||
	    (CtlP->rocketModel.model == MODEL_RP4M))
		return TRUE;
	else
		return FALSE;
}

static int cardIsRM3(CONTROLLER_T * CtlP)
{
	if ((CtlP->rocketModel.model == MODEL_UPCI_RM3_8PORT) ||
	    (CtlP->rocketModel.model == MODEL_UPCI_RM3_4PORT))
		return TRUE;
	else
		return FALSE;
}

static int cardIsRM4(CONTROLLER_T * CtlP)
{
	if ((CtlP->rocketModel.model == MODEL_UPCI_RM4_8PORT) ||
	    (CtlP->rocketModel.model == MODEL_UPCI_RM4_4PORT))
		return TRUE;
	else
		return FALSE;
}

static int cardIsRocketportPlus(CONTROLLER_T * CtlP)
{
	if ((CtlP->rocketModel.model == MODEL_RPP4) ||
	    (CtlP->rocketModel.model == MODEL_RPP8) ||
	    (CtlP->rocketModel.model == MODEL_RP2_232) ||
	    (CtlP->rocketModel.model == MODEL_RP2_422) ||
	    (CtlP->rocketModel.model == MODEL_UPCI_RPP8) ||
	    (CtlP->rocketModel.model == MODEL_UPCI_RPP2_232) ||
	    (CtlP->rocketModel.model == MODEL_UPCI_RPP2_422) ||
	    (CtlP->rocketModel.model == MODEL_UPCI_RPP4J) ||
	    (CtlP->rocketModel.model == MODEL_UPCI_RPP8J) ||
	    (CtlP->rocketModel.model == MODEL_UPCI_RPP8_422))
		return TRUE;
	else
		return FALSE;
}

static int cardIsURPPluswRI(CONTROLLER_T * CtlP)
{
	if ((CtlP->rocketModel.model == MODEL_UPCI_RPP8) ||
	    (CtlP->rocketModel.model == MODEL_UPCI_RPP2_232))
		return TRUE;
	else
		return FALSE;
}

/* *INDENT-OFF* */
/*  Comtrol Model Info Array - Every device and it's attributes */
modelInfo_t modelInfo[] = {
	//   model               model String                          numAiops    numPorts
	{MODEL_RP32INTF,         "RP 32 port w/external I/F",                4,      32},
	{MODEL_RP8INTF,          "RP 8 port w/external I/F",                 1,       8},
	{MODEL_RP16INTF,         "RP 16 port w/external I/F",                2,      16},
	{MODEL_RP8OCTA,          "RP 8 port w/octa cable",                   1,       8},
	{MODEL_RP4QUAD,          "RP 4 port w/quad cable",                   1,       4},
	{MODEL_RP8J,             "RP 8 port w/RJ11 connectors",              1,       8},
	{MODEL_RP4J,             "RP 4 port w/RJ45 connectors",              1,       4},
	{MODEL_RP8SNI,           "RP 8 port w/ DB78 SNI connector",          1,       8},
	{MODEL_RP16SNI,          "RP 16 port w/ DB78 SNI connector",         2,      16},
	{MODEL_RPP4,             "RP Plus 4 port",                           1,       4},
	{MODEL_RPP8,             "RP Plus 8 port",                           2,       8},
	{MODEL_RP2_232,          "RP Plus 2 port RS232",                     1,       2},
	{MODEL_RP2_422,          "RP Plus 2 port RS422",                     1,       2},
	{MODEL_RP6M,             "RM 6 port",                                1,       6},
	{MODEL_RP4M,             "RM 4 port",                                1,       4},
	{MODEL_UPCI_RP32INTF,    "RP UPCI 32 port w/external I/F",           4,      32},
	{MODEL_UPCI_RP8INTF,     "RP UPCI 8 port w/external I/F",            1,       8},
	{MODEL_UPCI_RP16INTF,    "RP UPCI 16 port w/external I/F",           2,      16},
	{MODEL_UPCI_RP8OCTA,     "RP UPCI 8 port w/octa cable",              1,       8},
	{MODEL_UPCI_RPP8,        "RP UPCI PLUS 8 port",                      2,       8},
	{MODEL_UPCI_RM3_8PORT,   "RP UPCI Rocketmodem III 8 port",           1,       8},
	{MODEL_UPCI_RM3_4PORT,   "RP UPCI Rocketmodem III 4 port",           1,       4},
	{MODEL_UPCI_RM4_8PORT,   "RP UPCI Rocketmodem IV 8 port",            1,       8},
	{MODEL_UPCI_RM4_4PORT,   "RP UPCI Rocketmodem IV 4 port",            1,       4},
	{MODEL_UPCI_RPP2_232,    "RP UPCI Plus 2 port RS232",                1,       2},
	{MODEL_UPCI_RPP2_422,    "RP UPCI Plus 2 port RS422",                1,       2},
	{MODEL_UPCI_RPP4J,       "RP UPCI Plus 4 port w/RJ45 connectors",    1,       4},
	{MODEL_UPCI_RPP8J,       "RP UPCI Plus 8 port w/RJ11 connectors",    2,       8},
	{MODEL_UPCI_RPP8_422,    "RP UPCI Plus 8 port RS422",                2,       8},
	{MODEL_CPCI_RP16INTF,    "RP Compact PCI 16 port w/external I/F",    2,      16},
	{MODEL_ISA,              "RP ISA",                                   1,       8}

};
/* *INDENT-ON* */

static int GetNumberOfModels(void)
{
	return (sizeof(modelInfo) / sizeof(modelInfo_t));
}

/*
 *  Gets the modelInfo struct corresponding to the model type
 *  Inputs:   Model type (PCI Model ID)
 *  Returns:  modelInfo_t pointer, or NULL if not found
 */
static modelInfo_t *GetModelInfo(int model)
{
	int i;

	for (i = 0; i < GetNumberOfModels(); i++) {
		if (modelInfo[i].model == model)
			return &modelInfo[i];
	}
	printk(KERN_WARNING "RocketPort: skipping Comtrol PCI card with unknown model number == 0x%08x\n", model);
	return NULL;
}

/*
 *  Sets the clock prescale value.
 *  Inputs:   CtlP pointer to controller struct, aiop number, prescale value.
 */
static void SetClockPrescale(CONTROLLER_T * CtlP, int aiopnum,
			     unsigned char prescale)
{
	sOutW(CtlP->AiopIO[aiopnum] + _INDX_ADDR, _CLK_PRE);
	sOutB(CtlP->AiopIO[aiopnum] + _INDX_DATA, prescale);
}
